@@ -22,6 +22,7 @@ namespace System.Drawing.Toolkit
 using System;
 using System.Runtime.InteropServices;
 using System.Collections;
+using DotGNU.Images;
 
 internal abstract class DrawingWindow : IToolkitWindow
 {
@@ -30,8 +31,8 @@ internal abstract class DrawingWindow : IToolkitWindow
 	protected internal IntPtr hwnd;
 	protected IntPtr hdc;
 	protected internal DrawingWindow parent;
-	// The window at the top of the heirarchy
-	protected internal DrawingWindow topOfHeirarchy;
+	// The window at the top of the hierarchy
+	protected internal DrawingWindow topOfhierarchy;
 	private Color background;
 	//cached when we create the window
 	private IntPtr backgroundBrush = IntPtr.Zero;
@@ -60,54 +61,47 @@ internal abstract class DrawingWindow : IToolkitWindow
 		this.toolkit = toolkit;
 	}
 
-	// Set the window title (top-level windows only).
-	void IToolkitWindow.SetTitle(String title)
-	{
-		if (hwnd == IntPtr.Zero)
-			throw new ApplicationException("DrawingWindow.SetTitle ERROR:Cant set title. Hwnd not created yet");
-		if(title == null)
-		{
-			title = String.Empty;
-		}
-		Win32.Api.SetWindowTextA(hwnd, title);
-		//Console.WriteLine("DrawingWindow.SetTitle, " + sink);
-	}
-
+	// Move this window underneath the last window with the same parent.
 	void IToolkitWindow.Lower()
 	{
-		//if (hwnd == IntPtr.Zero)
-		//	//Console.WriteLine("DrawingWindow.Lower ERROR: Cant lower window. Hwnd not created yet");
+		// Find the last window with the same parent as this.
+		DrawingWindow toWindow = null;
+		for (int to = toolkit.windowCount - 1; to >= 0 ; to--)
 		{
-			Win32.Api.SetWindowPos(hwnd, Win32.Api.SetWindowsPosPosition.HWND_BOTTOM, 0, 0, 0, 0, Win32.Api.SetWindowsPosFlags.SWP_NOMOVE | Win32.Api.SetWindowsPosFlags.SWP_NOSIZE);
-			//Console.WriteLine("DrawingWindow.Lower, " + sink);
+			toWindow = toolkit.windows[to];
+			if (parent.IsParentOf(toWindow))
+				break;
 		}
+		(this as IToolkitWindow).MoveToBelow(toWindow as IToolkitWindow);
+		//Console.WriteLine("DrawingWindow.Lower, " + sink);
 	}
 
+	// Move this window to the topmost position amongst the children.
 	void IToolkitWindow.Raise()
 	{
-		if (hwnd == IntPtr.Zero)
-			throw new ApplicationException("DrawingWindow.Raise ERROR: Cant raise window. Hwnd not created yet");
-		
-		Win32.Api.SetWindowPos(hwnd, Win32.Api.SetWindowsPosPosition.HWND_TOP, 0, 0, 0, 0, Win32.Api.SetWindowsPosFlags.SWP_NOMOVE | Win32.Api.SetWindowsPosFlags.SWP_NOSIZE);
+		MoveWindowTo(this, parent);
+		// Make the same changes to win32.
+		Win32.Api.SetWindowPos(hwnd, Win32.Api.SetWindowsPosPosition.HWND_TOP, 0, 0, 0, 0, Win32.Api.SetWindowsPosFlags.SWP_NOMOVE | Win32.Api.SetWindowsPosFlags.SWP_NOSIZE | Win32.Api.SetWindowsPosFlags.SWP_NOACTIVATE);
 		//Console.WriteLine("DrawingWindow.Raise, " + sink);
 	}
 
 	// Move this window to below one of its siblings.
-	// Is this really useful?
 	void IToolkitWindow.MoveToBelow(IToolkitWindow sibling)
 	{
+		DrawingWindow toWindow = sibling as DrawingWindow;
+		MoveWindowTo(this, toWindow);
+		// Make the same changes in win32.
+		Win32.Api.SetWindowPos(hwnd, toWindow.hwnd, 0, 0, 0, 0, Win32.Api.SetWindowsPosFlags.SWP_NOMOVE | Win32.Api.SetWindowsPosFlags.SWP_NOSIZE | Win32.Api.SetWindowsPosFlags.SWP_NOACTIVATE);
+		//Console.WriteLine("DrawingWindow.MoveToBelow, " + sink +", " + sibling);
 	}
 
 	// Move this window to above one of its siblings.
 	void IToolkitWindow.MoveToAbove(IToolkitWindow sibling)
 	{
-		if (hwnd == IntPtr.Zero)
-			throw new ApplicationException("DrawingWindow.MoveToAbove ERROR: Cant MoveToAbove. Hwnd not created yet");
-		Win32.Api.SetWindowPos(hwnd, (sibling as DrawingWindow).hwnd, 0, 0, 0, 0, Win32.Api.SetWindowsPosFlags.SWP_NOMOVE | Win32.Api.SetWindowsPosFlags.SWP_NOSIZE);
+		throw new NotSupportedException();
 		//Console.WriteLine("DrawingWindow.MoveToAbove, "+sink);
 	}
 
-
 	void IToolkitWindow.Reparent(IToolkitWindow parent, int x, int y)
 	{
 		throw new NotSupportedException("TopLevelWindow can't have a parent");
@@ -174,9 +168,15 @@ internal abstract class DrawingWindow : IToolkitWindow
 		{
 			//Console.WriteLine("DrawingWindow.Capture, "+ value +", "+ sink);
 			if (value)
+			{
 				toolkit.capturedWindow  = this;
+				Win32.Api.SetCapture(hwnd);
+			}
 			else
+			{
 				toolkit.capturedWindow  = null;
+				Win32.Api.ReleaseCapture();
+			}
 		}
 	}
 
@@ -188,15 +188,27 @@ internal abstract class DrawingWindow : IToolkitWindow
 	}
 
 	//Destroy window
-	public virtual void Destroy()
+	void IToolkitWindow.Destroy()
 	{
 		// Call will ignore if hwnd = IntPtr.Zero
 		Win32.Api.DestroyWindow(hwnd);
 		//Console.WriteLine("DrawingWindow.Destroy, "+sink);
 	}
 
-	void IToolkitWindow.SetWindowFlags(System.Drawing.Toolkit.ToolkitWindowFlags flags)
+	public void Dispose()
 	{
+		Dispose(true);
+		GC.SuppressFinalize(this);
+	}
+
+	protected virtual void Dispose(bool disposing)
+	{
+		(this as IToolkitWindow).Destroy();
+	}
+
+	~DrawingWindow()
+	{
+		Dispose(false);
 	}
 
 	// Set the background of the window to a solid color.
@@ -222,7 +234,7 @@ internal abstract class DrawingWindow : IToolkitWindow
 			int leftAdjust, topAdjust, rightAdjust, bottomAdjust;
 			Toolkit.GetWindowAdjust(out leftAdjust, out topAdjust, out rightAdjust, out bottomAdjust, flags);
 			outsideDimensions = new Rectangle( x - leftAdjust, y - topAdjust, width + leftAdjust + rightAdjust, height + topAdjust + bottomAdjust);
-			Win32.Api.SetWindowPos(hwnd, Win32.Api.SetWindowsPosPosition.HWND_TOP, outsideDimensions.Left, outsideDimensions.Top, outsideDimensions.Width, outsideDimensions.Height, Win32.Api.SetWindowsPosFlags.SWP_NOSENDCHANGING);
+			Win32.Api.SetWindowPos(hwnd, 0, outsideDimensions.Left, outsideDimensions.Top, outsideDimensions.Width, outsideDimensions.Height, Win32.Api.SetWindowsPosFlags.SWP_NOSENDCHANGING | Win32.Api.SetWindowsPosFlags.SWP_NOACTIVATE | Win32.Api.SetWindowsPosFlags.SWP_NOZORDER);
 		}
 		//Console.WriteLine("DrawingWindow.MoveResize, " + sink +",["+x+","+y+","+width+","+height+"]");
 	}
@@ -232,15 +244,13 @@ internal abstract class DrawingWindow : IToolkitWindow
 	{
 		get
 		{
-			if (hwnd == IntPtr.Zero)
-				throw new ApplicationException("can not get visible state, control is not created yet");
-			return Win32.Api.IsWindowVisible(hwnd);
+			return visible;
 		}
 		set
 		{
 			visible = value;
 			if (hwnd != IntPtr.Zero)
-				setVisible();
+				SetVisible();
 			//Console.WriteLine("DrawingWindow.setIsMapped "+sink+",visible="+value);
 		}
 	}
@@ -274,16 +284,6 @@ internal abstract class DrawingWindow : IToolkitWindow
 	{
 	}
 
-	void IToolkitWindow.Iconify()
-	{
-		if (hwnd == IntPtr.Zero)
-			throw new ApplicationException("Can not minimize, window not created yet");
-		
-		Win32.Api.CloseWindow(hwnd);
-		//Console.WriteLine("DrawingWindow.Iconify, "+sink);
-	}
-
-
 	// Get a toolkit graphics object for this window.
 	IToolkitGraphics IToolkitWindow.GetGraphics()
 	{
@@ -299,7 +299,7 @@ internal abstract class DrawingWindow : IToolkitWindow
 	{
 		get
 		{
-			return (IToolkitWindow)parent;
+			return parent as IToolkitWindow;
 		}
 	}
 
@@ -308,6 +308,15 @@ internal abstract class DrawingWindow : IToolkitWindow
 		return hwnd;
 	}
 
+	// Set the cursor.  The toolkit may ignore "frame" if it already
+	// has a system-defined association for "cursorType".  Setting
+	// "cursorType" to "ToolkitCursorType.InheritParent" will reset
+	// the cursor to be the same as the parent window's.
+	void IToolkitWindow.SetCursor(ToolkitCursorType cursorType, Frame frame)
+	{
+		// TODO
+	}
+
 	// Invalidate a rectangle within this window.
 	public void Invalidate(int x, int y, int width, int height)
 	{
@@ -317,10 +326,9 @@ internal abstract class DrawingWindow : IToolkitWindow
 		Win32.Api.RECT r;
 		r.left = x;
 		r.top = y;
-		//TODO: Check if its +1
-		r.right = x + width + 1;
-		r.bottom = y + height + 1;
-		Win32.Api.InvalidateRect(hwnd, ref r, false);
+		r.right = x + width;
+		r.bottom = y + height;
+		Win32.Api.InvalidateRect(hwnd, ref r, true);
 		//Console.WriteLine("DrawingWindow.Invalidate, "+sink + " ["+x+","+y+","+width+","+height+"]");
 	}
 
@@ -328,35 +336,33 @@ internal abstract class DrawingWindow : IToolkitWindow
 	public void Invalidate()
 	{
 		if (hwnd != IntPtr.Zero)
-			Win32.Api.InvalidateRect(hwnd, IntPtr.Zero, false);
-	}
-
-	// Called when Windows wants to erase the form background. Use the provided hdc
-	internal void EraseBackground(IntPtr hdc) 
-	{
-		if (hwnd == IntPtr.Zero)
-			throw new ApplicationException("Can not EraseBackground, window not created yet");
-		
-		if (backgroundBrush!=IntPtr.Zero) 
-		{
-			Win32.Api.RECT clientRectangle;
-			Win32.Api.GetClientRect(hwnd, out clientRectangle);
-			clientRectangle = new System.Drawing.Win32.Api.RECT(clientRectangle.left, clientRectangle.top, clientRectangle.right, clientRectangle.bottom);
-			Win32.Api.FillRect(hdc, ref clientRectangle,backgroundBrush);
-		}
+			Win32.Api.InvalidateRect(hwnd, IntPtr.Zero, true);
+		//Console.WriteLine("DrawingWindow.Invalidate, "+sink);
 	}
 
 	// This occurs when a top level window (form) receives focus)
-	internal void Activate(int wParam, int lParam)
+	internal virtual int Activate(int wParam, int lParam)
 	{
+		//Console.WriteLine("DrawingWindow.Activate, " + sink + "," + ((Win32.Api.ActivateState)(wParam & 0xFF)).ToString());
 		if ((Win32.Api.ActivateState)(wParam & 0xFF) == Win32.Api.ActivateState.WA_INACTIVE)
-			return;
-		//Console.WriteLine("DrawingWindow.Activate, " + sink);
+		{
+			// Make sure we still dont hold the capture.
+			CaptureTopLevel(null);
+		}
 		// We must handle this ourselves
 		//Win32.Api.SetCursor(Win32.Api.LoadCursorA(IntPtr.Zero, Win32.Api.CursorName.IDC_ARROW));
-		//CaptureTopLevel = this;
+		return 0;
 	}
 
+	internal virtual int MouseActivate(DrawingWindow activateWindow)
+	{
+		//Console.WriteLine("DrawingWindow.MouseActivate, " + sink +"," + activateWindow.sink);
+		if (activateWindow is DrawingPopupWindow)
+			return (int)Win32.Api.WM_MOUSEACTIVATEReturn.MA_NOACTIVATE;
+		return (int)Win32.Api.WM_MOUSEACTIVATEReturn.MA_ACTIVATE;
+	}
+
+	// Set the capture to the top most parent of a window, or release the capture.
 	private void CaptureTopLevel(DrawingWindow window)
 	{
 			if (window == null)
@@ -368,12 +374,12 @@ internal abstract class DrawingWindow : IToolkitWindow
 					toolkit.capturedTopWindow = null;
 				}
 			}
-			else if (toolkit.capturedTopWindow != window.topOfHeirarchy)
+			else if (toolkit.capturedTopWindow != window.topOfhierarchy)
 			{
 				Win32.Api.ReleaseCapture();
-				Win32.Api.SetCapture(window.topOfHeirarchy.hwnd);
-				//Console.WriteLine("CaptureTopLevel, Switching " + toolkit.capturedTopWindow + " to " + window.topOfHeirarchy);
-				toolkit.capturedTopWindow = window.topOfHeirarchy;
+				Win32.Api.SetCapture(window.topOfhierarchy.hwnd);
+				//Console.WriteLine("CaptureTopLevel, Switching " + toolkit.capturedTopWindow + " to " + window.topOfhierarchy);
+				toolkit.capturedTopWindow = window.topOfhierarchy;
 			}
 	}
 
@@ -387,33 +393,20 @@ internal abstract class DrawingWindow : IToolkitWindow
 		
 		DrawingWindow actual;
 		DrawingWindow found = FindDeepestChild(lParam, out x, out y, out actual);
+		// Do Leave and Entered events.
 		if (toolkit.enteredWindow != actual)
 		{
 			if (toolkit.enteredWindow != null)
 			{
 				//Console.WriteLine("DrawingWindow.MouseLeave, " + toolkit.enteredWindow.sink);
-				// Check to see if the event changes the capture state if so redo
-				DrawingWindow prevCapturedWindow = toolkit.capturedWindow;
 				toolkit.enteredWindow.sink.ToolkitMouseLeave();
 				toolkit.enteredWindow = null;
-				if (prevCapturedWindow != toolkit.capturedWindow)
-				{
-					MouseMove(msg, wParam, lParam);
-					return;
-				}
 			}
 			if (actual != null)
 			{
 				//Console.WriteLine("DrawingWindow.MouseEnter, " + actual.sink);
 				toolkit.enteredWindow = actual;
-				// Check to see if the event changes the capture state if so redo
-				DrawingWindow prevCapturedWindow = toolkit.capturedWindow;
 				actual.sink.ToolkitMouseEnter();
-				if (prevCapturedWindow != toolkit.capturedWindow)
-				{
-					MouseMove(msg, wParam, lParam);
-					return;
-				}
 			}
 		}
 		
@@ -424,54 +417,57 @@ internal abstract class DrawingWindow : IToolkitWindow
 		// If this is the first time over a window, set the capture and try again
 		if (toolkit.capturedTopWindow == null)
 		{
-			if (toolkit.capturedWindow != null)
-				CaptureTopLevel(toolkit.capturedWindow);
-			else
+			if (toolkit.capturedWindow == null)
+			{
 				CaptureTopLevel(this);
-			MouseMove(msg, wParam, lParam);
+				MouseMove(msg, wParam, lParam);
+			}
+			else
+				CaptureTopLevel(toolkit.capturedWindow);
 			return;
 		}
-		
 
 		Win32.Api.POINT mouse;
 		mouse.x= MouseX(lParam) + DimensionsScreen.X;
 		mouse.y = MouseY(lParam) + DimensionsScreen.Y;
 		DrawingWindow actualWindow = toolkit.DrawingWindow(Win32.Api.WindowFromPoint(mouse));
 
-		if (toolkit.capturedWindow != null)
-			toolkit.capturedWindow.sink.ToolkitMouseMove(buttons, keys,0 ,x ,y, 0);
-		else if (actualWindow != null && actualWindow.topOfHeirarchy == toolkit.capturedTopWindow)
+		if (actualWindow != null && actualWindow.topOfhierarchy == toolkit.capturedTopWindow)
 		{
 			// On the window decorations.
 			if (found == null)
 				CaptureTopLevel(null);
 			else
 				// Send the message to the client window
-				found.sink.ToolkitMouseMove(buttons, keys,0 ,x ,y, 0);
+				found.sink.ToolkitMouseMove(buttons, keys, 0 ,x ,y, 0);
+			//Console.WriteLine("DrawingWindow.MouseMove, " + (actual==null?"null":actual.sink.ToString()) +", (" + x +", " + y +"), " + buttons +"," + keys );
 		}
 		else
 		{
-			CaptureTopLevel(null);
+			if (toolkit.capturedWindow == null)
+				CaptureTopLevel(null);
+			else
+				found.sink.ToolkitMouseMove(buttons, keys,0 ,x ,y, 0);
 			// ?Send the last mouse move beyond the window (Win NT+ behaviour)
-			//found.sink.ToolkitMouseMove(buttons, keys,0 ,x ,y, 0);
-			/*if (actualWindow != null)
-			{
-				MouseMove(msg, wParam, lParam);
-			}*/
-			// We should replay this mouse move
-			/*
-			// Convert this windows coordinates to the new top windows coordinates.
-			int x1 = MouseX(lParam) + DimensionsScreen.X - top.dimensions.X;
-			int y1 = MouseY(lParam) + DimensionsScreen.Y - top.dimensions.Y;
-			// Now repost this message to the top level window
-			lParam = x1 | y1 << 16;
-
-			// Send the message to the client window
-			Win32.Api.PostMessageA(hwnd, (Win32.Api.WindowsMessages)msg, wParam, lParam);
-			return;
-			*/
 		}
-		
+	}
+
+	// Need to look at z-order so we can't do this ourselves yet..
+	private DrawingWindow WindowFromPoint(int x, int y)
+	{
+		for (int i = toolkit.windowCount - 1; i >= 0; i--)
+		{
+			DrawingWindow w = toolkit.windows[i];
+			DrawingTopLevelWindow top = w as DrawingTopLevelWindow;
+			if (top != null)
+			{
+				if (top.outsideDimensions.Contains(x,y))
+					return w;
+			}
+			else if (w.DimensionsScreen.Contains(x, y))
+				return w;
+		}
+		return null;
 	}
 	
 	// Returns the dimensions relative to the screen.
@@ -511,6 +507,8 @@ internal abstract class DrawingWindow : IToolkitWindow
 			found.sink.ToolkitMouseDown(MapToToolkitMouseButtons(wParam), MapMouseToToolkitKeys(wParam), 1 ,x ,y ,0);
 			//Console.WriteLine("DrawingWindow.ButtonDown " + found.sink + " [" + x + "," + y + "], key:" + MapMouseToToolkitKeys(wParam) + ", button:" + MapToToolkitMouseButtons(wParam));
 		}
+		else
+			CaptureTopLevel(null);
 	}
 
 	internal void ButtonUp(int msg, int wParam, int lParam)
@@ -537,7 +535,7 @@ internal abstract class DrawingWindow : IToolkitWindow
 		}
 	}
 
-	// Returns the bottom child window that is underneath x, y and has a top level window of capturedTopWindow or capturedWindow (if some window has captured the most). The window must be visible and toplevelwindows decorations are excluded. Returns x and y relative to the found windows coordinates.
+	// Returns the bottom child window that is underneath x, y and that has a top level window of capturedTopWindow or capturedWindow (if some window has captured the mouse). The window must be visible and toplevelwindows decorations are excluded. Returns x and y relative to the found windows coordinates.
 	private DrawingWindow FindDeepestChild(int lParam, out int x, out int y, out DrawingWindow actualFound)
 	{
 		// Convert the mouse to screen coordinates
@@ -547,25 +545,36 @@ internal abstract class DrawingWindow : IToolkitWindow
 
 		DrawingWindow found = null;
 
-		// Child windows lower down the heirachy will always be last in the windows list
-		for (int i = 0; i < toolkit.windows.Count; i++)
+		// Child windows lower down the hierarchy will always be last in the windows list
+		for (int i = toolkit.windowCount - 1; i >= 0; i--)
 		{
-			DrawingWindow w = toolkit.windows[i] as DrawingWindow;
-			if (toolkit.capturedWindow != null)
+			DrawingWindow w = toolkit.windows[i];
+			if (!w.visible || !w.Visible)
+				continue;
+			if (toolkit.capturedWindow == null)
 			{
-				if (w.Visible && toolkit.capturedWindow.IsParentOf(w) && w.DimensionsScreen.Contains(x,y))
-					found = w;
-			}
-			else if (w.Visible && toolkit.capturedTopWindow == w.topOfHeirarchy)
-			{
-				// This will occur when the mouse is over the titlebar of a toplevel window.
-				if (w is DrawingTopLevelWindow && w.outsideDimensions.Contains(x,y) && !w.dimensions.Contains(x, y))
+				if (toolkit.capturedTopWindow == w.topOfhierarchy)
 				{
-					actualFound = null;
-					return null;
+					// This will occur when the mouse is over the titlebar of a toplevel window.
+					if (w.DimensionsScreen.Contains(x, y))
+					{
+						found = w;
+						break;
+					}
+					else if (w is DrawingTopLevelWindow && w.outsideDimensions.Contains(x,y))
+					{
+						actualFound = null;
+						return null;
+					}
 				}
-				else if (w.DimensionsScreen.Contains(x, y))
+			}
+			else
+			{
+				if (toolkit.capturedWindow.IsParentOf(w) && w.DimensionsScreen.Contains(x,y))
+				{
 					found = w;
+					break;
+				}
 			}
 		}
 
@@ -577,15 +586,10 @@ internal abstract class DrawingWindow : IToolkitWindow
 		if (found != null)
 		{
 			// Adjust the coordinates relative to the "found" window.
-			Rectangle child;
-			if (toolkit.capturedWindow == null)
-				child = found.DimensionsScreen;
-			else
-				child = toolkit.capturedWindow.DimensionsScreen;
+			Rectangle child = found.DimensionsScreen;
 			x -= child.X;
 			y -= child.Y;
 		}
-
 		return found;
 	}
 
@@ -631,8 +635,9 @@ internal abstract class DrawingWindow : IToolkitWindow
 			gr.SetClip(Rectangle.FromLTRB(myPS.rcPaintLeft, myPS.rcPaintTop, myPS.rcPaintRight, myPS.rcPaintBottom));
 			sink.ToolkitExpose( gr );
 			gr.Dispose();
+			//Console.WriteLine( "DrawingWindow.Paint "+ sink +","+gr.ClipBounds.ToString());
+
 		}
-		//Console.WriteLine( "DrawingWindow.Paint "+ sink);
 
 		Win32.Api.EndPaint( hwnd, ref myPS );
 	}
@@ -678,7 +683,28 @@ internal abstract class DrawingWindow : IToolkitWindow
 			{
 				int leftAdjust, topAdjust, rightAdjust, bottomAdjust;
 				Toolkit.GetWindowAdjust(out leftAdjust, out topAdjust, out rightAdjust, out bottomAdjust, flags);
-				sink.ToolkitExternalResize( pos.cx - leftAdjust - rightAdjust, pos.cy - topAdjust - bottomAdjust );
+				int width = pos.cx;
+				int height = pos.cy;
+				DrawingTopLevelWindow form = this as DrawingTopLevelWindow;
+				// Restrict the size if necessary.
+				if (form != null)
+				{
+					if (form.minimumSize != Size.Empty)
+					{
+						if (width < form.minimumSize.Width)
+							width = form.minimumSize.Width;
+						if (height < form.minimumSize.Height)
+							height = form.minimumSize.Height;
+					}
+					if (form.maximumSize != Size.Empty)
+					{
+						if (width > form.maximumSize.Width)
+							width = form.maximumSize.Width;
+						if (height > form.maximumSize.Height)
+							height = form.maximumSize.Height;
+					}
+				}
+				sink.ToolkitExternalResize( width - leftAdjust - rightAdjust, height - topAdjust - bottomAdjust );
 			}
 
 			// Now prevent windows from changing the position or size, System.Windows.Control will do that
@@ -766,7 +792,7 @@ internal abstract class DrawingWindow : IToolkitWindow
 	}
 
 	// This is called to set whether a window is visible or not. Can only happen once the window is created.
-	protected void setVisible()
+	protected void SetVisible()
 	{
 		if (visible)
 			Win32.Api.ShowWindow(hwnd,Win32.Api.ShowWindowCommand.SW_SHOWNA);
@@ -776,29 +802,94 @@ internal abstract class DrawingWindow : IToolkitWindow
 
 	abstract internal void CreateWindow();
 
-	// Make sure the parent window is before the child in the heirarchy
-	protected void AdjustWindows(DrawingWindow parent)
+	// Set the child right next to the parent in the heirarchy sure the parent window is before the child.
+	// Set the correct TopOfHierarchy for all the childwindows.
+	protected internal void ReparentOwn(DrawingWindow parent)
 	{
-		// Find all children windows and set their new top of heirarchy
-		DrawingWindow newTopOfHeirarchy = parent == null ? this : (parent as DrawingWindow).topOfHeirarchy;
-		foreach(DrawingWindow w in toolkit.windows)
-			if (this.IsParentOf(w))
-				w.topOfHeirarchy = newTopOfHeirarchy;
-		this.parent = parent as DrawingWindow;
-		// Make sure the parent window is before the child in the heirarchy
-		ArrayList windows = toolkit.windows;
-		int iChild = windows.IndexOf(this);
-		for(DrawingWindow w = parent; w != null; w = w.parent)
+		SetTopOfHierarchy();
+		this.parent = parent;
+		if (parent != null)
+			MoveWindowTo(this, parent);
+	}
+
+	protected void SetTopOfHierarchy()
+	{
+		// Find all children windows and set their new top of hierarchy
+		DrawingWindow newTopOfhierarchy;
+		if (parent == null)
+			newTopOfhierarchy = this;
+		else
+			newTopOfhierarchy = (parent as DrawingWindow).topOfhierarchy;
+		for (int i = 0; i < toolkit.windowCount; i++)
 		{
-			int pos = windows.IndexOf(w);
-			if (pos > iChild)
-			{
-				windows.Remove(w);
-				windows.Insert(iChild,w);
-			}
-			else
+			DrawingWindow w = toolkit.windows[i];
+			if (this.IsParentOf(w))
+				w.topOfhierarchy = newTopOfhierarchy;
+		}
+	}
+
+	// Move the window at position from to position to
+	// Also move all the children in the from windows hierarchy
+	protected void MoveWindowTo(DrawingWindow fromWindow, DrawingWindow toWindow)
+	{
+		// Find the from position.
+		int from;
+		for (from = 0; from < toolkit.windowCount; from++)
+		{
+			if (toolkit.windows[from] == fromWindow)
 				break;
 		}
+		// Find the to position.
+		int to;
+		for (to = 0; to < toolkit.windowCount; to++)
+		{
+			if (toolkit.windows[to] == toWindow)
+				break;
+		}
+		if (from == to + 1 || from == to)
+			return;
+		DrawingWindow[] move = new DrawingWindow[16];
+		int nextMove = 0;
+		// Move all children of "win" into the move array.
+		// Mark the children with null if they have been moved.
+		for (int i = from; i < toolkit.windowCount; i++)
+		{
+			DrawingWindow current = toolkit.windows[i];
+			if (fromWindow.IsParentOf(current) || current == fromWindow)
+			{
+				if (nextMove == move.Length)
+				{
+					DrawingWindow[] newMove = new DrawingWindow[move.Length * 2];
+					move.CopyTo(newMove, 0);
+					move = newMove;
+				}
+				toolkit.windows[i] = null;
+				move[nextMove++] = current;
+				if (i <= to)
+					to--;
+			}
+		}
+
+		// Fill in the null "holes".
+		int nonNullItem = from;
+		for(int i = from; i < toolkit.windowCount - nextMove; i++)
+		{
+			DrawingWindow window = toolkit.windows[i];
+			// Make sure we fill it with the next non null window.
+			for(;; nonNullItem++)
+			{
+				if (toolkit.windows[nonNullItem] != null)
+				{
+					toolkit.windows[i] = toolkit.windows[nonNullItem++];
+					break;
+				}
+			}
+		}
+
+		// Make room for the new windows.
+		Array.Copy(toolkit.windows,to + 1, toolkit.windows, to + 1 + nextMove, toolkit.windowCount - 1 - to - nextMove);
+		// Copy the move entries after the to location.
+		Array.Copy(move, 0, toolkit.windows , to + 1, nextMove);
 	}
 
 	// Returns whether a DrawingWindow is the parent of this window or is the window
