@@ -46,6 +46,7 @@ internal class Api
 
 	public enum WindowsMessages : int 
 		{
+			WM_NULL = 0x0,
 			WM_CREATE = 0x1,
 			WM_DESTROY = 0x2,
 			WM_ACTIVATE = 0x6,
@@ -57,7 +58,9 @@ internal class Api
 			WM_ERASEBKGND = 0x14,
 			WM_SETTINGCHANGE = 0x1A,
 			WM_SETCURSOR = 0x20,
+			WM_MOUSEACTIVATE = 0x21,
 			WM_WINDOWPOSCHANGING = 0x46,
+			WM_SETICON = 0x80,
 			WM_NCCALCSIZE = 0x83,
 			WM_NCHITTEST = 0x84,
 			WM_NCPAINT = 0x0085,
@@ -225,7 +228,8 @@ internal class Api
 		SW_MAXIMIZE = 3,
 		SW_SHOW = 5,
 		SW_MINIMIZE = 6,
-		SW_SHOWNA = 8 //Visible but no activate
+		SW_SHOWNA = 8, //Visible but no activate
+		SW_RESTORE = 9
 	}
 
 	public enum SystemCommand
@@ -381,6 +385,14 @@ internal class Api
 		BI_PNG = 5
 	}
 
+	public enum WM_MOUSEACTIVATEReturn
+	{
+		MA_ACTIVATE = 1,
+		MA_ACTIVATEANDEAT = 2,
+		MA_NOACTIVATE = 3,
+		MA_NOACTIVATEANDEAT = 4
+	}
+
 	[StructLayout(LayoutKind.Sequential,CharSet=CharSet.Ansi)]
 	public struct WNDCLASS 
 	{
@@ -669,25 +681,6 @@ internal class Api
 		public uint flags;
 	}
 
-// Not used	
-	[StructLayout(LayoutKind.Sequential)]
-	public class BITMAPINFO
-	{
-		public int biSize;
-		public int biWidth;
-		public int biHeight;
-		public short biPlanes;
-		public short biBitCount;
-		public BitMapInfoCompressionType biCompression;
-		public int biSizeImage = 0;
-		public int biXPelsPerMeter = 0;
-		public int biYPelsPerMeter = 0;
-		public int biClrUsed = 0;
-		public int biClrImportant = 0;
-		[MarshalAs(System.Runtime.InteropServices.UnmanagedType.ByValArray, SizeConst=1024)]
-		public byte[] bmiColors; 
-	};
-
 	[StructLayout(LayoutKind.Sequential)]
 	public struct RGNDATA
 	{
@@ -699,6 +692,32 @@ internal class Api
 		public byte[] buffer;
 	}
 
+	[StructLayout(LayoutKind.Sequential)]
+	public struct BITMAPV4INFO
+	{
+		public uint biSize; 
+		public int biWidth; 
+		public int biHeight; 
+		public ushort biPlanes; 
+		public ushort biBitCount; 
+		public uint biCompression; 
+		public uint biSizeImage; 
+		public int biXPelsPerMeter; 
+		public int biYPelsPerMeter; 
+		public uint biClrUsed; 
+		public uint biClrImportant; 
+	}
+
+	[StructLayout(LayoutKind.Sequential)]
+	public struct ICONINFO
+	{
+		public bool fIcon;
+		public uint xHotspot;
+		public uint yHotspot;
+		public IntPtr hbmMask;
+		public IntPtr hbmColor;
+	}
+
 	public delegate void TimerProc(IntPtr hwnd, uint uMsg, uint idEvent, uint dwTime);
 
 	[DllImport("user32")] //ANSI
@@ -791,9 +810,15 @@ internal class Api
 	[DllImport("gdi32")]
 	public static extern IntPtr CreateBrushIndirect(ref LOGBRUSH lplb);
 
+	[DllImport("gdi32")]
+	public static extern IntPtr CreatePatternBrush(IntPtr hbmp);
+
 	[DllImport("user32")] //ANSI
 	public static extern bool PostMessageA(IntPtr hwnd, WindowsMessages Msg, int wParam, int lParam);
 
+	[DllImport("user32")] //ANSI
+	public static extern int SendMessageA( IntPtr hWnd, WindowsMessages Msg, int wParam, IntPtr lParam);
+
 	[DllImport("user32")]
 	public static extern int GetSystemMetrics (SystemMetricsType nIndex);
 
@@ -815,7 +840,7 @@ internal class Api
 
 	//Measure size and width of text
 	[DllImport("gdi32")] //ANSI
-	public static extern int GetTextExtentPoint32A(IntPtr hdc, string str, int len, ref SIZE size);
+	public static extern int GetTextExtentPoint32A(IntPtr hdc, string str, int len, out SIZE size);
 	
 	[DllImport("gdi32")]
 	public static extern IntPtr GetStockObject( StockObjectType fnObject );
@@ -868,6 +893,9 @@ internal class Api
 	[DllImport("gdi32")]
 	public static extern bool Pie( IntPtr hdc, int nLeftRect,int nTopRect, int nRightRect, int nBottomRect, int nXRadial1, int nYRadial1, int nXRadial2, int nYRadial2 );
 
+	[DllImport("gdi32")]
+	public static extern bool Ellipse( IntPtr hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect );
+
 	[DllImport("gdi32")]
 	public static extern IntPtr SetTextColor( IntPtr hdc, int crColor);
 
@@ -878,10 +906,10 @@ internal class Api
 	public static extern bool IsWindowVisible( IntPtr hWnd);
 
 	[DllImport("gdi32")]
-	public static extern IntPtr CreatePatternBrush( IntPtr hbmp );
+	public static extern IntPtr CreateBitmap( int nWidth, int nHeight, uint cPlanes, uint cBitsPerPel, byte[] lpvBits);
 
 	[DllImport("gdi32")]
-	public static extern IntPtr CreateBitmap( int nWidth, int nHeight, uint cPlanes, uint cBitsPerPel, byte[] lpvBits);
+	public static extern IntPtr CreateDIBitmap( IntPtr hdc, byte[] lpbmih, uint fdwInit, byte[] lpbInit, byte[] lpbmi, uint fuUsage);
 
 	[DllImport("user32")]
 	public static extern int SetWindowLongA( IntPtr hWnd, SetWindowLongType nIndex, WindowStyle dwNewLong);
@@ -959,7 +987,13 @@ internal class Api
 	public static extern IntPtr WindowFromPoint( POINT Point);
 
 	[DllImport("gdi32")]
-	public static extern IntPtr CreateDIBSection(IntPtr hdc, [In,MarshalAs(UnmanagedType.LPStruct)] BITMAPINFO pbmi, DibColorTableType iUsage, out IntPtr ppvBits, IntPtr hSection, uint dwOffset);
+	public static extern IntPtr CreateCompatibleBitmap( IntPtr hdc, int nWidth, int nHeight);
+
+	[DllImport("gdi32")]
+	public static extern int SetDIBits( IntPtr hdc, IntPtr hbmp, uint uStartScan, uint cScanLines, ref byte lpvBits, byte[] lpbmi, DibColorTableType fuColorUse);
+
+	[DllImport("gdi32")]
+	public static extern int SetDIBitsToDevice( IntPtr hdc, int XDest, int YDest, uint dwWidth, uint dwHeight, int XSrc, int YSrc, uint uStartScan, uint cScanLines, ref byte lpvBits, byte[] lpbmi, DibColorTableType fuColorUse);
 
 	[DllImport("gdi32")]
 	public static extern int BitBlt (IntPtr hdcDest, int x, int y, int nWidth, int nHeight, IntPtr hSrcDC, int xSrc, int ySrc, RopType dwRop);
@@ -967,22 +1001,6 @@ internal class Api
 	[DllImport("gdi32")]
 	public static extern IntPtr CreateCompatibleDC( IntPtr hdc);
 
-	[DllImport("gdi32")]
-	public static extern int SetDIBitsToDevice(
-		IntPtr hdc,                 // handle to DC
-		int XDest,               // x-coord of destination upper-left corner
-		int YDest,               // y-coord of destination upper-left corner 
-		uint dwWidth,           // source rectangle width
-		uint dwHeight,          // source rectangle height
-		int XSrc,                // x-coord of source lower-left corner
-		int YSrc,                // y-coord of source lower-left corner
-		uint uStartScan,         // first scan line in array
-		uint cScanLines,         // number of scan lines
-		ref byte lpvBits,     // array of DIB bits
-		BITMAPINFO lpbmi, // bitmap information
-	uint fuColorUse          // RGB or palette indexes
-		);
-
 	[DllImport("kernel32")]
 	public static extern IntPtr GlobalLock( IntPtr hMem);
 
@@ -1001,6 +1019,11 @@ internal class Api
 	[DllImport("gdi32")]
 	public static extern uint GetRegionData( IntPtr hRgn, uint dwCount, ref byte lpRgnData);
 
+	[DllImport("user32")]
+	public static extern IntPtr CreateIcon( IntPtr hInstance, int nWidth, int nHeight, byte cPlanes, byte cBitsPixel, ref byte lpbANDbits, ref byte lpbXORbits);
+
+	[DllImport("user32")]
+	public static extern IntPtr CreateIconIndirect(ref ICONINFO piconinfo);
 
 }//Api
 
