@@ -25,6 +25,7 @@ using System;
 using System.Drawing;
 using System.Drawing.Drawing2D;
 using System.Drawing.Toolkit;
+using DotGNU.Images;
 using Xsharp;
 
 internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
@@ -34,6 +35,7 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 	private IToolkitEventSink sink;
 	private ButtonName button;
 	private bool hasCapture;
+	private static Xsharp.Cursor[] cursors;
 
 	// Constructor.
 	public DrawingWindow(IToolkit toolkit, Widget parent,
@@ -143,12 +145,6 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 				Lower();
 			}
 
-	// Iconify the window.
-	void IToolkitWindow.Iconify()
-			{
-				// Not used for ordinary windows.
-			}
-
 	// Reparent this window to underneath a new parent.
 	void IToolkitWindow.Reparent(IToolkitWindow parent, int x, int y)
 			{
@@ -169,12 +165,6 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 					(toolkit, new Xsharp.Graphics(this));
 			}
 
-	// Set the window title (top-level windows only).
-	void IToolkitWindow.SetTitle(String title)
-			{
-				// Not used for ordinary windows.
-			}
-
 	// Set the foreground of the window to a solid color.
 	void IToolkitWindow.SetForeground(System.Drawing.Color color)
 			{
@@ -187,12 +177,6 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 				Background = DrawingToolkit.DrawingToXColor(color);
 			}
 
-	// Change the set of supported window decorations and functions.
-	void IToolkitWindow.SetWindowFlags(ToolkitWindowFlags flags)
-			{
-				// Not used for ordinary windows.
-			}
-
 	// Move this window to above one of its siblings.
 	void IToolkitWindow.MoveToAbove(IToolkitWindow sibling)
 			{
@@ -220,17 +204,27 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 	// Invalidate a rectangle within this window.
 	void IToolkitWindow.Invalidate(int x, int y, int width, int height)
 			{
-				DrawingToolkit.ValidateWindowPosition(ref x, ref y);
+				DrawingToolkit.ValidateWindowPositionPaint(ref x, ref y);
 				DrawingToolkit.ValidateWindowSize(ref width, ref height);
-				Repaint(x, y, width, height);
+				Repaint(x, y, width + 1, height + 1);
 			}
 
 	// Force an update of all invalidated regions.
 	void IToolkitWindow.Update()
 			{
+				Update();
 				Display.Flush();
 			}
 
+	// Set the cursor.  The toolkit may ignore "frame" if it already
+	// has a system-defined association for "cursorType".  Setting
+	// "cursorType" to "ToolkitCursorType.InheritParent" will reset
+	// the cursor to be the same as the parent window's.
+	void IToolkitWindow.SetCursor(ToolkitCursorType cursorType, Frame frame)
+			{
+				ModifyCursor(this, cursorType, frame);
+			}
+
 	// Map an Xsharp key description into a "ToolkitKeys" value.
 	private static ToolkitKeys MapKey(KeyName key)
 			{
@@ -239,6 +233,7 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 					case KeyName.XK_BackSpace:		return ToolkitKeys.Back;
 					case KeyName.XK_Tab:			return ToolkitKeys.Tab;
 					case KeyName.XK_KP_Tab:			return ToolkitKeys.Tab;
+					case KeyName.XK_ISO_Left_Tab:	return ToolkitKeys.Tab;
 					case KeyName.XK_Linefeed:		return ToolkitKeys.LineFeed;
 					case KeyName.XK_Clear:			return ToolkitKeys.Clear;
 					case KeyName.XK_Return:			return ToolkitKeys.Enter;
@@ -617,6 +612,136 @@ internal sealed class DrawingWindow : InputOutputWidget, IToolkitWindow
 				return newRegion;
 			}
 
+	// Modify the cursor on a widget.
+	internal static void ModifyCursor
+				(Widget widget, ToolkitCursorType cursorType, Frame frame)
+			{
+				if(cursorType == ToolkitCursorType.InheritParent)
+				{
+					// Change the cursor back to the same value as the parent.
+					widget.Cursor = null;
+				}
+				else if(cursorType == ToolkitCursorType.Default)
+				{
+					// Try to inherit the desktop's cursor for the default.
+					Widget parent = widget.Parent;
+					while(parent != null)
+					{
+						if(parent.Cursor != null)
+						{
+							break;
+						}
+						parent = parent.Parent;
+					}
+					if(parent != null)
+					{
+						widget.Cursor = new Cursor(CursorType.XC_left_ptr);
+					}
+					else
+					{
+						widget.Cursor =
+							new Cursor(CursorType.XC_inherit_parent);
+					}
+				}
+				else
+				{
+					// Create a cursor based on the supplied image.
+					if(frame != null &&
+					   frame.PixelFormat ==
+					   		DotGNU.Images.PixelFormat.Format1bppIndexed &&
+					   frame.Mask != null)
+					{
+						if(cursorType != ToolkitCursorType.UserDefined)
+						{
+							lock(typeof(DrawingWindow))
+							{
+								if(cursors == null)
+								{
+									cursors = new Xsharp.Cursor [32];
+								}
+								Xsharp.Cursor curs = cursors[(int)cursorType];
+								if(curs == null)
+								{
+									cursors[(int)cursorType] = curs =
+										new Xsharp.Cursor
+											(widget.Screen, frame);
+								}
+								widget.Cursor = curs;
+							}
+						}
+						else
+						{
+							widget.Cursor = new Xsharp.Cursor
+								(widget.Screen, frame);
+						}
+						return;
+					}
+
+					// Last ditch attempt - use the nearest X11 cursor.
+					CursorType type;
+					switch(cursorType)
+					{
+						case ToolkitCursorType.AppStarting:
+							type = CursorType.XC_watch;
+							break;
+						case ToolkitCursorType.Arrow:
+						case ToolkitCursorType.Default:
+							type = CursorType.XC_left_ptr;
+							break;
+						case ToolkitCursorType.Cross:
+							type = CursorType.XC_crosshair;
+							break;
+						case ToolkitCursorType.IBeam:
+							type = CursorType.XC_xterm;
+							break;
+						case ToolkitCursorType.No:
+							type = CursorType.XC_X_cursor;
+							break;
+						case ToolkitCursorType.SizeAll:
+							type = CursorType.XC_fleur;
+							break;
+						case ToolkitCursorType.SizeNS:
+						case ToolkitCursorType.VSplit:
+							type = CursorType.XC_sb_v_double_arrow;
+							break;
+						case ToolkitCursorType.SizeWE:
+						case ToolkitCursorType.HSplit:
+							type = CursorType.XC_sb_h_double_arrow;
+							break;
+						case ToolkitCursorType.UpArrow:
+							type = CursorType.XC_sb_up_arrow;
+							break;
+						case ToolkitCursorType.WaitCursor:
+							type = CursorType.XC_watch;
+							break;
+						case ToolkitCursorType.Help:
+							type = CursorType.XC_question_arrow;
+							break;
+						case ToolkitCursorType.Hand:
+							type = CursorType.XC_hand2;
+							break;
+						case ToolkitCursorType.SizeNWSE:
+						case ToolkitCursorType.SizeNESW:
+						case ToolkitCursorType.NoMove2D:
+						case ToolkitCursorType.NoMoveHoriz:
+						case ToolkitCursorType.NoMoveVert:
+						case ToolkitCursorType.PanEast:
+						case ToolkitCursorType.PanNE:
+						case ToolkitCursorType.PanNorth:
+						case ToolkitCursorType.PanNW:
+						case ToolkitCursorType.PanSE:
+						case ToolkitCursorType.PanSouth:
+						case ToolkitCursorType.PanSW:
+						case ToolkitCursorType.PanWest:
+						case ToolkitCursorType.UserDefined:
+						default:
+							type = CursorType.XC_circle;
+							break;
+					}
+					widget.Cursor = new Cursor(type);
+				}
+			}
+
 }; // class DrawingWindow
 
 }; // namespace System.Drawing.Toolkit
