@@ -3,6 +3,7 @@
  *			"System.Windows.Forms.Form" class.
  *
  * Copyright (C) 2003  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2003  Neil Cawse.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,10 +33,9 @@ public class Form : ContainerControl
 	private IButtonControl acceptButton;
 	private IButtonControl cancelButton;
 	private bool autoScale;
-	private bool isMdiChild;
-	private bool isMdiContainer;
 	private bool keyPreview;
 	private bool topLevel;
+	internal bool dialogResultIsSet;
 	private Size autoScaleBaseSize;
 	private DialogResult dialogResult;
 	private FormBorderStyle borderStyle;
@@ -57,6 +57,7 @@ public class Form : ContainerControl
 	internal static Form activeForm;
 	private bool showInTaskbar;
 	private bool controlBox;
+	private MdiClient mdiClient;
 	
 	// Constructor.
 	public Form()
@@ -74,6 +75,13 @@ public class Form : ContainerControl
 			}
 
 	// Get or set this control's properties.
+	private IToolkitTopLevelWindow ToolkitWindow
+			{
+				get
+				{
+					return (toolkitWindow as IToolkitTopLevelWindow);
+				}
+			}
 	public IButtonControl AcceptButton
 			{
 				get
@@ -92,13 +100,18 @@ public class Form : ContainerControl
 					return activeForm;
 				}
 			}
-	[TODO]
 	public Form ActiveMdiChild
 			{
 				get
 				{
-					// TODO
-					return null;
+					if(mdiClient != null)
+					{
+						return mdiClient.ActiveChild;
+					}
+					else
+					{
+						return null;
+					}
 				}
 			}
 	public bool AutoScale
@@ -211,6 +224,7 @@ public class Form : ContainerControl
 				set
 				{
 					dialogResult = value;
+					dialogResultIsSet = true;
 				}
 			}
 	public FormBorderStyle FormBorderStyle
@@ -239,7 +253,6 @@ public class Form : ContainerControl
 					SetWindowFlag(ToolkitWindowFlags.Help, value);
 				}
 			}
-	[TODO]
 	public Icon Icon
 			{
 				get
@@ -248,28 +261,37 @@ public class Form : ContainerControl
 				}
 				set
 				{
-					// TODO: send the icon to the window manager.
-					icon = value;
+					if(icon != value)
+					{
+						icon = value;
+						if(ToolkitWindow != null)
+						{
+							ToolkitWindow.SetIcon(value);
+						}
+					}
 				}
 			}
 	public bool IsMdiChild
 			{
 				get
 				{
-					return isMdiChild;
+					return (MdiParent != null);
 				}
 			}
-	[TODO]
 	public bool IsMdiContainer
 			{
 				get
 				{
-					return isMdiContainer;
+					return (mdiClient != null);
 				}
 				set
 				{
-					// TODO: switch the form into MDI mode
-					isMdiContainer = value;
+					if(value && mdiClient == null)
+					{
+						// Convert the form into MDI mode.
+						mdiClient = new MdiClient();
+						Controls.Add(mdiClient);
+					}
 				}
 			}
 	public bool KeyPreview
@@ -307,7 +329,6 @@ public class Form : ContainerControl
 					// TODO
 				}
 			}
-	[TODO]
 	public Size MaximumSize
 			{
 				get
@@ -316,7 +337,6 @@ public class Form : ContainerControl
 				}
 				set
 				{
-					// TODO: send the maximum size to the window manager.
 					if(value.Width < 0)
 					{
 						throw new ArgumentOutOfRangeException
@@ -327,17 +347,34 @@ public class Form : ContainerControl
 						throw new ArgumentOutOfRangeException
 							("value.Height", S._("SWF_NonNegative"));
 					}
-					maximumSize = value;
+					if(maximumSize != value)
+					{
+						maximumSize = value;
+						if(ToolkitWindow != null)
+						{
+							ToolkitWindow.SetMaximumSize(value);
+						}
+						if (maximumSize.Width > 0 && Width > maximumSize.Width)
+							Width = maximumSize.Width;
+						if (maximumSize.Height > 0 && Height > maximumSize.Height)
+							Height = maximumSize.Height;
+					}
 				}
 			}
 	public Form[] MdiChildren
 			{
 				get
 				{
-					return mdiChildren;
+					if(mdiClient != null)
+					{
+						return mdiClient.MdiChildren;
+					}
+					else
+					{
+						return new Form [0];
+					}
 				}
 			}
-	[TODO]
 	public Form MdiParent
 			{
 				get
@@ -346,7 +383,14 @@ public class Form : ContainerControl
 				}
 				set
 				{
-					// TODO: validate the parent
+					if(mdiParent != null || value == null || Created)
+					{
+						return;
+					}
+					if(value.mdiClient == null)
+					{
+						return;
+					}
 					mdiParent = value;
 				}
 			}
@@ -361,7 +405,6 @@ public class Form : ContainerControl
 					SetWindowFlag(ToolkitWindowFlags.Minimize, value);
 				}
 			}
-	[TODO]
 	public Size MinimumSize
 			{
 				get
@@ -370,7 +413,6 @@ public class Form : ContainerControl
 				}
 				set
 				{
-					// TODO: send the minimum size to the window manager.
 					if(value.Width < 0)
 					{
 						throw new ArgumentOutOfRangeException
@@ -381,7 +423,18 @@ public class Form : ContainerControl
 						throw new ArgumentOutOfRangeException
 							("value.Height", S._("SWF_NonNegative"));
 					}
-					minimumSize = value;
+					if(minimumSize != value)
+					{
+						minimumSize = value;
+						if(ToolkitWindow != null)
+						{
+							ToolkitWindow.SetMinimumSize(value);
+						}
+						if (minimumSize.Width > 0 && Width < minimumSize.Width)
+							Width = minimumSize.Width;
+						if (minimumSize.Height > 0 && Height < minimumSize.Height)
+							Height = minimumSize.Height;
+					}
 				}
 			}
 	public MainMenu Menu
@@ -407,14 +460,14 @@ public class Form : ContainerControl
 								other.Menu = null;
 							}
 						}
-						// Get the ClientSize before we add the menu
+						// Get the ClientSize before we add the menu.
 						Size clientSize = ClientSize;
 						menu = value;
 						if(menu != null)
 						{
 							menu.AddToForm(this);
 						}
-						// The clientsize must be the same as the old one
+						// The ClientSize must be the original.
 						ClientSize = clientSize;
 					}
 				}
@@ -547,7 +600,6 @@ public class Form : ContainerControl
 					transparencyKey = value;
 				}
 			}
-	[TODO]
 	public FormWindowState WindowState
 			{
 				get
@@ -556,25 +608,75 @@ public class Form : ContainerControl
 				}
 				set
 				{
-					// TODO: pass the state change to the window manager
-					windowState = value;
+					if(windowState != value)
+					{
+						windowState = value;
+						if(ToolkitWindow != null)
+						{
+							if(value == FormWindowState.Normal)
+							{
+								ToolkitWindow.Restore();
+							}
+						#if !CONFIG_COMPACT_FORMS
+							else if(value == FormWindowState.Minimized)
+							{
+								ToolkitWindow.Iconify();
+							}
+						#endif
+							else if(value == FormWindowState.Maximized)
+							{
+								ToolkitWindow.Maximize();
+							}
+						}
+					}
 				}
 			}
 
 	// Create the toolkit window underlying this control.
 	internal override IToolkitWindow CreateToolkitWindow(IToolkitWindow parent)
 			{
-				// TODO: dialog forms
 				CreateParams cp = CreateParams;
 
 				// Create the window and set its initial caption.
-				IToolkitWindow window =
-					ToolkitManager.Toolkit.CreateTopLevelWindow
-						(cp.Width - ToolkitDrawSize.Width, cp.Height - ToolkitDrawSize.Height, this);
+				IToolkitTopLevelWindow window;
+				if(mdiParent == null)
+				{
+					window = ToolkitManager.Toolkit.CreateTopLevelWindow
+						(cp.Width - ToolkitDrawSize.Width,
+						 cp.Height - ToolkitDrawSize.Height, this);
+				}
+				else
+				{
+					mdiParent.mdiClient.CreateControl();
+					IToolkitMdiClient mdi =
+						(mdiParent.mdiClient.toolkitWindow as
+							IToolkitMdiClient);
+					window = mdi.CreateChildWindow
+						(cp.X, cp.Y,
+						 cp.Width - ToolkitDrawSize.Width,
+						 cp.Height - ToolkitDrawSize.Height, this);
+				}
 				window.SetTitle(cp.Caption);
 
-				// Adjust the window decorations to match our requirements.
+				// Adjust the window hints to match our requirements.
 				SetWindowFlags(window);
+				if(icon != null)
+				{
+					window.SetIcon(icon);
+				}
+				window.SetMaximumSize(maximumSize);
+				window.SetMinimumSize(minimumSize);
+			#if !CONFIG_COMPACT_FORMS
+				if(windowState == FormWindowState.Minimized)
+				{
+					window.Iconify();
+				}
+				else
+			#endif
+				if(windowState == FormWindowState.Maximized)
+				{
+					window.Maximize();
+				}
 
 				// Center the window on-screen if necessary.
 				if(formStartPosition == FormStartPosition.CenterScreen)
@@ -583,7 +685,7 @@ public class Form : ContainerControl
 					window.MoveResize
 							((screenSize.Width - cp.Width) / 2,
 							 (screenSize.Height - cp.Height) / 2,
-							 cp.Width, cp.Height);
+							 window.Dimensions.Width, window.Dimensions.Height);
 				}
 				return window;
 			}
@@ -619,7 +721,6 @@ public class Form : ContainerControl
 					break;
 
 					case FormBorderStyle.Fixed3D:
-					case FormBorderStyle.FixedDialog:
 					case FormBorderStyle.FixedSingle:
 					{
 						flags &= ~(ToolkitWindowFlags.Maximize |
@@ -628,6 +729,15 @@ public class Form : ContainerControl
 					}
 					break;
 
+					case FormBorderStyle.FixedDialog:
+					{
+						flags &= ~(ToolkitWindowFlags.Maximize |
+								   ToolkitWindowFlags.ResizeHandles |
+								   ToolkitWindowFlags.Resize);
+						flags |= ToolkitWindowFlags.Dialog;
+					}
+					break;
+
 					case FormBorderStyle.FixedToolWindow:
 					{
 						flags &= ~(ToolkitWindowFlags.Maximize |
@@ -647,13 +757,17 @@ public class Form : ContainerControl
 					}
 					break;
 				}
+				if((flags & ToolkitWindowFlags.Modal) != 0)
+				{
+					flags |= ToolkitWindowFlags.Dialog;
+				}
 				return flags;
 			}
 
 	// Set the current state of the window decoration flags on a window.
 	private void SetWindowFlags(IToolkitWindow window)
 			{
-				window.SetWindowFlags(GetFullFlags());
+				((IToolkitTopLevelWindow)window).SetWindowFlags(GetFullFlags());
 			}
 
 	// Set the current state of a window decoration flag.
@@ -706,10 +820,12 @@ public class Form : ContainerControl
 			}
 
 	// Layout the MDI children of this form.
-	[TODO]
 	public void LayoutMdi(MdiLayout value)
 			{
-				// TODO
+				if(mdiClient != null)
+				{
+					mdiClient.LayoutMdi(value);
+				}
 			}
 
 	// Remove an owned form from this form.
@@ -741,11 +857,57 @@ public class Form : ContainerControl
 #endif // !CONFIG_COMPACT_FORMS
 
 	// Show this form as a modal dialog.
-	[TODO]
 	private DialogResult ShowDialog(Form owner)
 			{
-				// TODO
-				return DialogResult.None;
+				// Bail out if this dialog is already displayed modally.
+				if(Modal)
+				{
+					return DialogResult.None;
+				}
+
+				// Reset the dialog result.
+				dialogResult = DialogResult.None;
+				dialogResultIsSet = false;
+
+				// Mark this form as modal.
+				SetWindowFlag(ToolkitWindowFlags.Modal, true);
+				try
+				{
+					// Create the control.  We must do this before
+					// we set the owner or make the form visible.
+					CreateControl();
+
+					// Set the dialog owner.
+					IToolkitTopLevelWindow toolkitWindow = ToolkitWindow;
+					if(toolkitWindow != null)
+					{
+						if(owner != null)
+						{
+							toolkitWindow.SetDialogOwner(owner.ToolkitWindow);
+						}
+						else
+						{
+							toolkitWindow.SetDialogOwner(null);
+						}
+					}
+
+					// Make the form visible.
+					Visible = true;
+
+					// Enter a message loop until the dialog result is set.
+					Application.InnerMessageLoop(this);
+				}
+				finally
+				{
+					// Make sure that the form is not visible.
+					Visible = false;
+
+					// The form is no longer modal.
+					SetWindowFlag(ToolkitWindowFlags.Modal, false);
+				}
+
+				// Return the dialog result.
+				return dialogResult;
 			}
 	public DialogResult ShowDialog()
 			{
@@ -1168,9 +1330,9 @@ public class Form : ContainerControl
 	// Override the "TextChanged" event.
 	protected override void OnTextChanged(EventArgs e)
 			{
-				if(toolkitWindow != null)
+				if(ToolkitWindow != null)
 				{
-					toolkitWindow.SetTitle(Text);
+					ToolkitWindow.SetTitle(Text);
 				}
 				base.OnTextChanged(e);
 			}
@@ -1182,18 +1344,38 @@ public class Form : ContainerControl
 			}
 
 	// Process a command key.
-	[TODO]
 	protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
 			{
-				// TODO
-				return base.ProcessCmdKey(ref msg, keyData);
+				if (base.ProcessCmdKey(ref msg, keyData))
+					return true;
+				if (menu != null)
+					return menu.ProcessCmdKey(ref msg, keyData);
+				return false;
 			}
 
 	// Process a dialog key.
-	[TODO]
 	protected override bool ProcessDialogKey(Keys keyData)
 			{
-				// TODO
+				if ((keyData & (Keys.Control | Keys.Alt)) == 0)
+				{
+					Keys key = keyData & Keys.KeyCode;
+					if (key == Keys.Return)
+					{
+						if (acceptButton != null)
+						{
+							acceptButton.PerformClick();
+							return true;
+						}
+					}
+					else if (key == Keys.Escape)
+					{
+						if (cancelButton != null)
+						{
+							cancelButton.PerformClick();
+							return true;
+						}
+					}
+				}
 				return base.ProcessDialogKey(keyData);
 			}
 
@@ -1212,11 +1394,16 @@ public class Form : ContainerControl
 			}
 
 	// Process the tab key.
-	[TODO]
 	protected override bool ProcessTabKey(bool forward)
 			{
-				// TODO
-				return base.ProcessTabKey(forward);
+				return SelectNextControl(ActiveControl, forward, true, true, true);
+			}
+
+	protected override bool ProcessDialogChar(char charCode)
+			{
+				if (GetTopLevel() && ProcessMnemonic(charCode))
+					return true; 
+				return base.ProcessDialogChar(charCode); 
 			}
 
 	// Inner core of "Scale".
@@ -1228,13 +1415,37 @@ public class Form : ContainerControl
 			}
 
 	// Select this control.
-	[TODO]
 	protected override void Select(bool directed, bool forward)
 			{
-				// TODO
-				base.Select(directed, forward);
+				if (directed)
+					base.SelectNextControl(null, forward, true, true, false);
+			
+				if (TopLevel)
+					toolkitWindow.Focus();
+
+				Form parent = ParentForm;
+				if (parent != null)
+					parent.ActiveControl = this;
 			}
 
+	protected override void UpdateDefaultButton()
+			{
+				// Find the bottom active control.
+				Control c = this;
+				while (c is ContainerControl)
+					c = (c as ContainerControl).ActiveControl;
+
+				if (c is IButtonControl)
+				{
+					// Notify the previous button that it is not the default.
+					if (acceptButton != null)
+						(acceptButton as IButtonControl).NotifyDefault(false);
+					acceptButton = c as IButtonControl;
+					if (acceptButton != null)
+						(acceptButton as IButtonControl).NotifyDefault(true);
+				}
+			} 
+
 	// Inner core of "SetBounds".
 	protected override void SetBoundsCore
 				(int x, int y, int width, int height,
@@ -1309,7 +1520,21 @@ public class Form : ContainerControl
 				}
 			}
 
-	
+	// Window state change request received.
+	internal override void WindowStateChanged(FormWindowState state)
+			{
+				windowState = state;
+			}
+
+	// Event that is called when an MDI child is activated or deactivated.
+	internal override void MdiActivate(IToolkitWindow child)
+			{
+				if(mdiClient != null)
+				{
+					mdiClient.Activate(child);
+				}
+				OnMdiChildActivate(EventArgs.Empty);
+			}
 
 #if !CONFIG_COMPACT_FORMS
 
