@@ -67,6 +67,7 @@ public abstract class ScrollBar : Control
 		idleTimer = new Timer();
 		idleTimer.Tick += new EventHandler(idleTimer_Tick);
 		idleTimer.Interval = 1;
+		BackColor = SystemColors.ScrollBar;
 	}
 
 	// Properties
@@ -75,9 +76,9 @@ public abstract class ScrollBar : Control
 		get { return base.BackColor; }
 		set
 		{
-			if (value == base.BackColor) { return; }
+			if (value == base.backColor) { return; }
 			base.BackColor = value;
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -88,7 +89,7 @@ public abstract class ScrollBar : Control
 		{
 			if (value == base.BackgroundImage) { return; }
 			base.BackgroundImage = value;
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -105,8 +106,7 @@ public abstract class ScrollBar : Control
 		{
 			if (value == decDown) { return; }
 			decDown = value;
-			using (Graphics g = CreateGraphics())
-				Draw(g, decrement);
+			Invalidate(decrement);
 			if (value)
 			{
 				timer.Tick += new EventHandler(Decrement);
@@ -137,9 +137,9 @@ public abstract class ScrollBar : Control
 		get { return base.ForeColor; }
 		set
 		{
-			if (value == base.ForeColor) { return; }
+			if (value == base.foreColor) { return; }
 			base.ForeColor = value;
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -157,8 +157,7 @@ public abstract class ScrollBar : Control
 		{
 			if (value == incDown) { return; }
 			incDown = value;
-			using (Graphics g = CreateGraphics())
-				Draw(g, increment);
+			Invalidate(increment);
 			if (value)
 			{
 				timer.Tick += new EventHandler(Increment);
@@ -192,7 +191,7 @@ public abstract class ScrollBar : Control
 				largeChange = maximum - minimum + 1;
 			
 			LayoutScrollBar();
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -211,7 +210,7 @@ public abstract class ScrollBar : Control
 			if (value > maximum)
 				Value = maximum;
 			LayoutScrollBar();
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -230,7 +229,7 @@ public abstract class ScrollBar : Control
 			if (value < minimum)
 				Value = minimum;
 			LayoutScrollBar();
-			Draw();
+			Invalidate();
 		}
 	}
 
@@ -319,13 +318,6 @@ public abstract class ScrollBar : Control
 		RedrawTrackBar(oldBounds);
 	}
 
-	// Create a graphics and redraw the entire scroll bar
-	private void Draw()
-	{
-		using (Graphics g = CreateGraphics())
-			Draw(g, ClientRectangle);
-	}
-
 	// Draw if visible and created
 	private void Draw(Graphics g, Rectangle drawBounds)
 	{
@@ -462,7 +454,7 @@ public abstract class ScrollBar : Control
 
 	protected override void OnEnabledChanged(EventArgs e)
 	{
-		Draw();
+		Invalidate();
 		base.OnEnabledChanged(e);
 	}
 
@@ -641,14 +633,12 @@ public abstract class ScrollBar : Control
 		if (incDown)
 		{
 			IncrementDown = false;
-			using(Graphics g = CreateGraphics())
-				Draw(g, increment);
+			Invalidate(increment);
 		}
 		else if (decDown)
 		{
 			DecrementDown = false;
-			using(Graphics g = CreateGraphics())
-				Draw(g, decrement);
+			Invalidate(decrement);
 		}
 		else if (trackDown)
 		{
@@ -749,16 +739,9 @@ public abstract class ScrollBar : Control
 	// When the trackbar moves, to prevent flickering, we only redraw what we have to
 	private void RedrawTrackBar( Rectangle oldBounds)
 	{
-		// Exclude any area in the oldPos that will be drawn by the newPos
-		Region old = new Region(oldBounds);
-		old.Exclude(bar);
-		using (Graphics g = CreateGraphics())
-		{
-			// Replace old
-			Draw(g, Rectangle.Truncate( old.GetBounds(g)));
-			// Draw new
-			Draw(g, bar);
-		}
+		Region region = new Region(oldBounds);
+		region.Union(bar);
+		Invalidate(region);
 	}
 
 	protected override void OnMouseUp(MouseEventArgs e)
@@ -801,11 +784,30 @@ public abstract class ScrollBar : Control
 
 	protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
 	{
+		bool modified = (x != Left || y != Top || width != Width || height != Height);
 		base.SetBoundsCore (x, y, width, height, specified);
-		LayoutScrollBar();
-		Draw();
+		if (modified && IsHandleCreated)
+		{
+			LayoutScrollBar();
+			Invalidate();
+		}
 	}
 
+	protected override void OnCreateControl()
+	{
+		base.OnCreateControl ();
+		LayoutScrollBar();
+	}
+
+
+	protected override void OnVisibleChanged(EventArgs e)
+	{
+		base.OnVisibleChanged (e);
+		if (IsHandleCreated)
+			LayoutScrollBar();
+	}
+
+
 	// Generate the events when the scrollbar is incremented or decremented
 	private void GenerateManualScrollEvents(int newValue, ScrollEventType type)
 	{
