@@ -119,15 +119,20 @@ namespace System.Windows.Forms
 
 		protected virtual void OnSelectedIndexChanged(EventArgs e)
 		{
+			if (selectedIndex == -1 || selectedIndex >= TabCount)
+				return;
+			SuspendLayout();
+			positionInfo = null;
 			if (prevSelectedIndex > -1)
 			{
 				GetChildByIndex( prevSelectedIndex ).Visible = false;
 			}
 			GetChildByIndex( selectedIndex ).Visible = true;
 			
-			DrawTabs();
 			if (SelectedIndexChanged != null)
 				SelectedIndexChanged( this, EventArgs.Empty );
+			ResumeLayout();
+			InvalidateTabs();
 		}
 
 		protected void RemoveAll()
@@ -186,7 +191,7 @@ namespace System.Windows.Forms
 				{
 					hotTrackIndex = -1;
 					hotTrack = value;
-					DrawTabs();
+					InvalidateTabs();
 				}
 			}
 		}
@@ -200,7 +205,7 @@ namespace System.Windows.Forms
 			set
 			{
 				imageList = value;
-				DrawTabs();
+				InvalidateTabs();
 			}
 		}
 
@@ -336,61 +341,13 @@ namespace System.Windows.Forms
 
 		private void Draw(Graphics g)
 		{
-			Region clipMoveButtons = g.Clip.Clone();
+			// Exclude the moveButtons from the draw to prevent flicker.
 			if (moveButtonsShowing)
 			{
-				moveButtonLeftBounds = moveButtonRightBounds = new Rectangle( Width - moveButtonSize.Width *2, 1, moveButtonSize.Width, moveButtonSize.Height );
+				moveButtonLeftBounds = moveButtonRightBounds = new Rectangle( Width - moveButtonSize.Width *2, 1, moveButtonSize.Width, moveButtonSize.Height);
 				moveButtonRightBounds.Offset( moveButtonLeftBounds.Width, 0);
-				Region r = new Region(moveButtonLeftBounds);
-				r.Union(moveButtonRightBounds);
-				g.SetClip(r, Drawing.Drawing2D.CombineMode.Exclude);
-			}
-
-			Region clip = g.Clip.Clone();
-
-			Rectangle newClient = GetTabBaseBounds();
-			// Draw the tab edging
-			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.RaisedInner, Border3DSide.Left);
-			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.Raised, Border3DSide.Bottom);
-			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.Raised, Border3DSide.Right);
-			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.RaisedInner, Border3DSide.Top);
-			
-			// Clear out the space behind the tabs
-			using ( Brush background = new SolidBrush(BackColor) )
-				g.FillRectangle( background, GetTabsBounds());
-
-			// Draw each tab in the tabControl row 0 first, bottom row last, selected item very last
-			for( int row = 0; row < PositionInfo.totalRows; row++ )
-			{
-				for( int i = 0; i < tabPageCollection.Count; i++ )
-				{
-					if (row == PositionInfo.positions[i].row && i != selectedIndex)
-					{
-						Rectangle bounds = GetTabRect(i);
-						// Remove bottom line off bounds if not selected so border isnt covered
-						g.SetClip(new Rectangle(bounds.Left, bounds.Top, bounds.Width, bounds.Height -1), Drawing.Drawing2D.CombineMode.Intersect);
-						OnDrawItem( new DrawItemEventArgs(g, Font, bounds, i, DrawItemState.None, ForeColor, BackColor));
-						g.Clip = clip;
-					}
-				}
-			}
-			if (selectedIndex > -1)
-			{
-				Rectangle bounds = GetTabRect(selectedIndex);
-				g.SetClip(new Rectangle(bounds.Left, bounds.Top, bounds.Width, bounds.Height), Drawing.Drawing2D.CombineMode.Intersect);
-						
-				DrawItemState state = Focused?DrawItemState.Focus:DrawItemState.None;
-				OnDrawItem( new DrawItemEventArgs(g, Font, bounds, selectedIndex, state, ForeColor, BackColor));
-			}
-
-			// Draw the Tab move buttons if necessary
-			if (moveButtonsShowing)
-			{
-				
-
 				using (SolidBrush b = new SolidBrush( foreColor) )
 				{
-					g.Clip = clipMoveButtons;
 					ControlPaint.DrawButton(g, moveButtonLeftBounds, moveButtonLeftState);
 					// Left Arrow
 					g.FillPolygon(b, new Point[]
@@ -409,7 +366,54 @@ namespace System.Windows.Forms
 					});
 					
 				}
+				Region r = new Region(moveButtonLeftBounds);
+				r.Union(moveButtonRightBounds);
+				g.SetClip(r, Drawing.Drawing2D.CombineMode.Exclude);
 			}
+			Region clip = g.Clip.Clone();
+
+			Rectangle newClient = GetTabBaseBounds();
+			// Draw the tab edging
+			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.RaisedInner, Border3DSide.Left);
+			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.Raised, Border3DSide.Bottom);
+			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.Raised, Border3DSide.Right);
+			ControlPaint.DrawBorder3D(g, newClient, Border3DStyle.RaisedInner, Border3DSide.Top);
+
+			// Draw each tab in the tabControl row 0 first, bottom row last, selected item very last
+			for( int row = 0; row < PositionInfo.totalRows; row++ )
+			{
+				for( int i = 0; i < tabPageCollection.Count; i++ )
+				{
+					if (row == PositionInfo.positions[i].row && i != selectedIndex)
+					{
+						Rectangle bounds = GetTabRect(i);
+						// Remove bottom line off bounds if not selected so border isnt covered
+						g.SetClip(new Rectangle(bounds.Left, bounds.Top, bounds.Width, bounds.Height -1), Drawing.Drawing2D.CombineMode.Intersect);
+						OnDrawItem( new DrawItemEventArgs(g, Font, bounds, i, DrawItemState.None, ForeColor, BackColor));
+						g.Clip = clip;
+					}
+				}
+			}
+			if (selectedIndex > -1 && selectedIndex < TabPages.Count)
+			{
+				Rectangle bounds = GetTabRect(selectedIndex);
+				g.SetClip(new Rectangle(bounds.Left, bounds.Top, bounds.Width, bounds.Height), Drawing.Drawing2D.CombineMode.Intersect);
+				
+				DrawItemState state;
+				if (Focused)
+					state = DrawItemState.Focus;
+				else
+					state = DrawItemState.None;
+				OnDrawItem( new DrawItemEventArgs(g, Font, bounds, selectedIndex, state, ForeColor, BackColor));
+			}
+
+		}
+
+		public void DrawButton(int index)
+		{
+			if (index == -1 || index >= TabPages.Count)
+				return;
+			Invalidate(new Region( positionInfo.positions[index].bounds ));
 		}
 
 		public Rectangle GetTabRect( int index )
@@ -434,9 +438,9 @@ namespace System.Windows.Forms
 						// Find the width of a row
 						int rowWidth;
 						if (alignment == TabAlignment.Top || alignment == TabAlignment.Bottom)
-							rowWidth = Bounds.Width - indent * 2;
+							rowWidth = Width - indent * 2;
 						else
-							rowWidth = Bounds.Height - indent * 2;
+							rowWidth = Height - indent * 2;
 		
 						int maxRow = 0;
 						// Size the tabs based on the TabSizeMode
@@ -450,7 +454,7 @@ namespace System.Windows.Forms
 								SizeTabsFillRight ( ref tabs, graphics, rowWidth, out maxRow );
 						}
 						// Do we need to move the row that was selected to the bottom of the tabs?
-						if (selectedIndex > -1)
+						if (selectedIndex > -1 && selectedIndex < TabCount)
 						{
 							// Check to see if we have selected a tab that isnt on the last row and move the tab row down
 							if (tabs[selectedIndex].row != maxRow)
@@ -774,6 +778,10 @@ namespace System.Windows.Forms
 			}
 			set
 			{
+				if (value == -1)
+				{
+					value = 0;
+				}
 				if (value != selectedIndex)
 				{
 					prevSelectedIndex = selectedIndex;
@@ -838,16 +846,44 @@ namespace System.Windows.Forms
 			StringFormat format = new StringFormat();
 			format.Alignment = StringAlignment.Center;
 			format.LineAlignment = StringAlignment.Center;
+			format.FormatFlags = StringFormatFlags.NoWrap;
 			using (Brush brush = new SolidBrush( color )) 
 			{
 				graphics.DrawString( text, Font, brush, bounds, format );
 			}
 		}
 
+		protected override bool IsInputKey(Keys keyData)
+		{
+			if ((keyData & Keys.Alt) == 0)
+			{
+				Keys key = keyData & Keys.KeyCode;
+				if (key == Keys.Prior || key == Keys.Next || key == Keys.End || key == Keys.Home)
+					return true;
+			}
+			return base.IsInputKey(keyData); 
+		}
+
+		protected override void OnKeyDown(KeyEventArgs e)
+		{
+			// Ctrl Tab and Ctrl Shift Tab to cycle between tabs.
+			if (e.KeyCode == Keys.Tab && (e.KeyData & Keys.Control) != 0)
+			{
+				if (SelectedIndex != -1)
+				{
+					if ((e.KeyData & Keys.Shift) == 0)
+						SelectedIndex = (SelectedIndex + 1) % TabCount;
+					else
+						SelectedIndex = (SelectedIndex + TabCount - 1) % TabCount;
+					e.Handled = true;
+				}
+			}
+			base.OnKeyDown(e);
+		}
+
 		// Select the tab that is clicked
 		protected override void OnMouseDown(MouseEventArgs e)
 		{
-			base.OnMouseDown (e);
 			if (moveButtonsShowing && moveButtonLeftBounds.Contains( e.X, e.Y ))
 			{
 				moveButtonLeftState = ButtonState.Pushed;
@@ -855,7 +891,7 @@ namespace System.Windows.Forms
 				{
 					moveButtonsTabOffset--;
 					positionInfo = null;
-					DrawTabs();
+					InvalidateTabs();
 				}
 			}
 			else if (moveButtonsShowing && moveButtonRightBounds.Contains(e.X, e.Y))
@@ -865,19 +901,39 @@ namespace System.Windows.Forms
 				{
 					moveButtonsTabOffset++;
 					positionInfo = null;
-					DrawTabs();
+					InvalidateTabs();
 				}
 			}
 			else
 			{
 				moveButtonRightState = ButtonState.Normal;
-				Focus();
-				int selectedIndex =  GetMouseOverTab( e.X, e.Y );
-				if (selectedIndex > -1)
-					SelectedIndex = selectedIndex;
+				int newSelectedIndex =  GetMouseOverTab( e.X, e.Y );
+				if (newSelectedIndex > -1)
+				{
+					SelectedIndex = newSelectedIndex;
+					// Makes sure that when tabbing off the tab into the child controls that the tab order resets.
+					ActiveControl = this;
+				}
 			}
+			base.OnMouseDown (e);
+
 		}
 
+		protected override void OnLeave(EventArgs e)
+		{
+			// Redraw the selected button to clear the focus.
+			DrawButton(selectedIndex);
+			base.OnLeave (e);
+		}
+
+		protected override void OnEnter(EventArgs e)
+		{
+			// Redraw the selected button to set the focus.
+			DrawButton(selectedIndex);
+			base.OnEnter (e);
+		}
+
+
 		// Find which tab the mouse is over
 		private int GetMouseOverTab(int x, int y)
 		{
@@ -896,21 +952,12 @@ namespace System.Windows.Forms
 			if (moveButtonLeftState != ButtonState.Normal)
 			{
 				moveButtonLeftState = ButtonState.Normal;
-				using (Graphics g = CreateGraphics())
-				{
-					g.Clip = new Region(moveButtonLeftBounds);
-					Draw(g);
-				}
+				Invalidate(moveButtonLeftBounds);
 			}
 			if (moveButtonRightState != ButtonState.Normal)
 			{
 				moveButtonRightState = ButtonState.Normal;
-				using (Graphics g = CreateGraphics())
-				{
-					g.Clip = new Region(moveButtonRightBounds);
-					Draw(g);
-				}
-			
+				Invalidate(moveButtonRightBounds);
 			}
 			base.OnMouseUp (e);
 		}
@@ -927,20 +974,10 @@ namespace System.Windows.Forms
 					Region r = new Region(positionInfo.positions[hotTrackIndex].bounds);
 					if (prevHotTrackIndex > -1)
 						r.Union( positionInfo.positions[prevHotTrackIndex].bounds );
-					using (Graphics g = CreateGraphics())
-					{
-						g.Clip =r;
-						Draw(g);
-					}
+					Invalidate(r);
 				}
 				else if (hotTrackIndex == -1 && prevHotTrackIndex > -1)
-				{
-					using (Graphics g = CreateGraphics())
-					{
-						g.Clip = new Region( positionInfo.positions[prevHotTrackIndex].bounds );
-						Draw(g);
-					}
-				}
+					Invalidate(new Region( positionInfo.positions[prevHotTrackIndex].bounds ));
 
 			}
 			base.OnMouseMove (e);
@@ -952,11 +989,7 @@ namespace System.Windows.Forms
 			{
 				int prevHotTrackIndex = hotTrackIndex;
 				hotTrackIndex = -1;
-				using (Graphics g = CreateGraphics())
-				{
-					g.Clip = new Region( positionInfo.positions[prevHotTrackIndex].bounds );
-					Draw(g);
-				}
+				Invalidate( new Region( positionInfo.positions[prevHotTrackIndex].bounds ));
 			}
 			base.OnMouseLeave (e);
 		}
@@ -966,6 +999,8 @@ namespace System.Windows.Forms
 		{
 			get
 			{
+				if (toolkitWindow == null)
+					return ClientRectangle;
 				Rectangle rect = GetTabBaseBounds();
 				return new Rectangle( rect.Left + padding.X, rect.Top + padding.Y, rect.Width - 2* padding.X, rect.Height - 2 * padding.Y );
 			}
@@ -989,13 +1024,13 @@ namespace System.Windows.Forms
 			switch( alignment )
 			{
 				case TabAlignment.Left:
-					return new Rectangle( offset, 0, Bounds.Width - offset, Bounds.Height );
+					return new Rectangle( offset, 0, Width - offset, Height );
 				case TabAlignment.Top:
-					return new Rectangle( 0, offset, Bounds.Width, Bounds.Height - offset );
+					return new Rectangle( 0, offset, Width, Height - offset );
 				case TabAlignment.Right:
-					return new Rectangle( 0, 0, Bounds.Width - offset, Bounds.Height );
+					return new Rectangle( 0, 0, Width - offset, Height );
 				case TabAlignment.Bottom:
-					return new Rectangle( 0, 0, Bounds.Width, Bounds.Height - offset );
+					return new Rectangle( 0, 0, Width, Height - offset );
 				default:
 					return Rectangle.Empty;
 			}
@@ -1019,19 +1054,19 @@ namespace System.Windows.Forms
 			switch( alignment )
 			{
 				case TabAlignment.Left:
-					return new Rectangle( 0, 0, offset, Bounds.Height );
+					return new Rectangle( 0, 0, offset, Height );
 				case TabAlignment.Top:
-					return new Rectangle( 0, 0, Bounds.Width, offset );
+					return new Rectangle( 0, 0, Width, offset );
 				case TabAlignment.Right:
-					return new Rectangle( Bounds.Width - offset, 0, offset, Bounds.Height );
+					return new Rectangle( Width - offset, 0, offset, Height );
 				case TabAlignment.Bottom:
-					return new Rectangle( 0, Bounds.Height - offset, Bounds.Width, offset );
+					return new Rectangle( 0, Height - offset, Width, offset );
 				default:
 					return Rectangle.Empty;
 			}
 		}
 
-		internal void DrawTabs()
+		internal void InvalidateTabs()
 		{
 			if (!IsHandleCreated)
 				return;
@@ -1052,31 +1087,52 @@ namespace System.Windows.Forms
 					bounds =  new Rectangle(bounds.Left, bounds.Top - 1, bounds.Width, bounds.Height + 1);
 					break;
 			}
-			using (Graphics g = CreateGraphics())
-			{
-				g.SetClip(bounds);
-				Draw(g);
-			}
-		}
-
-		//HACK: This is a bug in Contol.cs. Remove this when its fixed
-		// One way or the other, PerformLayout should be called when a control is created
-		protected override void OnCreateControl()
-		{
-			base.OnCreateControl ();
-			PerformLayout();
+			Invalidate(bounds);
 		}
 
 		// The event thats called when controls need to relayout their contents
 		protected override void OnLayout(LayoutEventArgs e)
 		{
 			base.OnLayout (e);
-			// Force the position info to update
-			positionInfo = null;
-			// This forces the correct Bounds for all TabPages and is automatically called whenever the bounds could change eg. adding new tab
+			SetTabPageBounds();
+		}
+
+		protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
+		{
+			base.SetBoundsCore (x, y, width, height, specified);
+			if (!IsHandleCreated)
+			{
+				return;
+			}
+			// Force the positions of the tabs to update if the size changes
+			if ((specified & BoundsSpecified.Size) != 0)
+			{
+				positionInfo = null;
+				SetTabPageBounds();
+				Invalidate(false);
+			}
+		}
+
+		protected override void OnCreateControl()
+		{
+			base.OnCreateControl();
+			if (selectedIndex == -1)
+			{
+				SelectedIndex = 0;
+			}
+			SetTabPageBounds();
+		}
+
+		// This causes the child tab pages to have their bounds set to DisplayRectangle.
+		private void SetTabPageBounds()
+		{
 			foreach( TabPage tabPage in tabPageCollection)
-				tabPage.Bounds = Rectangle.Empty;
-			
+			{
+				if (tabPage.Visible)
+				{
+					tabPage.Bounds = Rectangle.Empty;
+				}
+			}
 		}
 
 	}
