@@ -194,7 +194,7 @@ public class TextBox : TextBoxBase
 			if (scrollBars == value)
 				return;
 			scrollBars = value;
-			SetupScrollBars();
+			layout = null;
 		}
 	}
 
@@ -276,7 +276,7 @@ public class TextBox : TextBoxBase
 		}
 
 
-		if (vScrollBar != null)
+		if (vScrollBar != null && vScrollBar.Visible)
 		{
 			vScrollBar.Bounds = new Rectangle(ClientRectangle.Width - vScrollBar.Width, 0, vScrollBar.Width, height);
 			
@@ -290,7 +290,7 @@ public class TextBox : TextBoxBase
 			vScrollBar.SmallChange = Font.Height;
 			vScrollBar.LargeChange = TextDrawArea.Height + 1;
 		}
-		if (hScrollBar != null)
+		if (hScrollBar != null && hScrollBar.Visible)
 		{
 			hScrollBar.Bounds = new Rectangle(0, ClientRectangle.Height - hScrollBar.Height, width, hScrollBar.Height);
 			int maximum = MaxTextDimensions.Width;
@@ -383,10 +383,12 @@ public class TextBox : TextBoxBase
 			SelectInternal(Text.Length, 0);
 			if (!inTextChangedEvent)
 			{
-				CaretSetEndSelection();
-
-				ResetView();
-				Redraw(ControlGraphics);
+				if (IsHandleCreated)
+				{
+					CaretSetEndSelection();
+					ResetView();
+					Redraw(ControlGraphics);
+				}
 				OnTextChanged(EventArgs.Empty);
 			}
 
@@ -451,11 +453,13 @@ public class TextBox : TextBoxBase
 			if(textAlign != value)
 			{
 				textAlign = value;
-				// Layout changes
-				LayoutFromText(Text);
-				SetScrollBarPositions();
-				
-				ResetView();
+				if (IsHandleCreated)
+				{
+					// Layout changes
+					LayoutFromText(Text);
+					SetScrollBarPositions();
+					ResetView();
+				}
 				OnTextAlignChanged(EventArgs.Empty);
 			}
 		}
@@ -616,6 +620,11 @@ public class TextBox : TextBoxBase
 	// Redraw a specific portion of the textbox
 	private void Redraw(Graphics g)
 	{
+		if (layout == null)
+		{
+			LayoutFromText(Text);
+			ResetView();
+		}
 		// Draw scrollbar corner if both are visible
 		if (vScrollBar != null && hScrollBar != null && vScrollBar.Visible && hScrollBar.Visible)
 			g.FillRectangle(SystemBrushes.Control, hScrollBar.Right, vScrollBar.Bottom, vScrollBar.Width, hScrollBar.Height);
@@ -645,11 +654,10 @@ public class TextBox : TextBoxBase
 	// Handle the event when multiline is changed.
 	private void HandleMultilineChanged(object sender, EventArgs e)
 	{
-		SetupScrollBars();
+		layout = null;
 		// Set back the actual chosen height
 		// Will cause LayoutFromText to be called
 		Height = chosenHeight;
-		CaretReset();
 	}
 
 	private Graphics ControlGraphics
@@ -743,12 +751,14 @@ public class TextBox : TextBoxBase
 	// All rendered in client coordinates.
 	protected void LayoutFromText(String newText)
 	{
+		if (!IsHandleCreated)
+			return;
 		if (layout == null)
 		{
 			layout = new LayoutInfo();
 			layout.Items = new LayoutInfo.Item[0];
 		}
-		
+
 		// Optimization - only re-layout from the beginning of the last line modified.
 		// Find posLine, the position of the beginning of the line we must start updating from.
 		// yLine is the y coordinate of this point.
@@ -792,20 +802,13 @@ public class TextBox : TextBoxBase
 		else if (textAlign == HorizontalAlignment.Center)
 			format.Alignment = StringAlignment.Center;
 	
-		// To use MeasureCharacterRanges, we need to specify that we want to measure each character individually.
-		// This MS api isnt very clever!
-		CharacterRange[] range = new CharacterRange[measureText.Length];
-		for (int i = 0; i < measureText.Length; i++)
-			range[i] = new CharacterRange(i, 1);
-		format.SetMeasurableCharacterRanges(range);
-		
-		Region[] bounds;
+		Rectangle[] bounds;
 		if (measureText.Length == 0)
-			bounds = new Region[0];
+			bounds = new Rectangle[0];
 		else if (passwordChar == 0)
-			bounds = ControlGraphics.MeasureCharacterRanges(measureText, Font, measureBounds, format);
+			bounds = ControlGraphics.MeasureCharacters(measureText, Font, measureBounds, format);
 		else
-			bounds = ControlGraphics.MeasureCharacterRanges(new string(passwordChar, measureText.Length), Font, measureBounds, format);
+			bounds = ControlGraphics.MeasureCharacters(new string(passwordChar, measureText.Length), Font, measureBounds, format);
 		LayoutInfo.Item[] newItems = new LayoutInfo.Item[newText.Length];
 		// Copy in the previously measured items.
 		Array.Copy(layout.Items, 0, newItems, 0, posLine);
@@ -825,7 +828,7 @@ public class TextBox : TextBoxBase
 		{
 			LayoutInfo.Item item = new LayoutInfo.Item();
 			char c = newText[i];
-			Rectangle rect = Rectangle.Truncate(bounds[i - posLine].GetBounds( ControlGraphics));
+			Rectangle rect = bounds[i - posLine];
 			if (c == '\r')
 			{
 				item.type = LayoutInfo.Item.CharType.CR;
@@ -857,6 +860,7 @@ public class TextBox : TextBoxBase
 			layout.Items[i] = item;
 			prevType = item.type;
 		}
+		SetupScrollBars();
 	}
 
 	// Make sure the caret is visible
@@ -936,6 +940,8 @@ public class TextBox : TextBoxBase
 	{
 		if (start == selectionStartActual && length == selectionLengthActual)
 			return;
+		if (!IsHandleCreated)
+			return;
 		Region newRegion = new Region(RectangleF.Empty);
 		selectionStartActual = start;
 		selectionLengthActual = length;
@@ -974,6 +980,11 @@ public class TextBox : TextBoxBase
 	// Called to change the text. Sets the update to whats needed to but doesnt change the selection point or caret
 	private void SetTextActual( string text)
 	{
+		if (!IsHandleCreated)
+		{
+			(this as Control).text = text;
+			return;
+		}
 		// Layout the new text. Compare with old layout, Creating a region for areas that must be updated.
 		bool prevLayout = layout != null;
 		LayoutInfo oldLayout = null;
@@ -1211,8 +1222,6 @@ public class TextBox : TextBoxBase
 			base.SetClientSizeCore (x - 2 * 2, y - 2 * 2);
 	}
 
-
-
 	protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
 	{
 		chosenHeight = height;
@@ -1220,6 +1229,8 @@ public class TextBox : TextBoxBase
 		if (!Multiline)
 			height = ClientToBounds(Size.Empty).Height + Font.Height + 1;
 		base.SetBoundsCore (x, y, width, height, specified);
+		if (!IsHandleCreated)
+			return;
 		// If the height or width changes then relayout the text
 		if ((specified & BoundsSpecified.Height) != 0 | (specified & BoundsSpecified.Width) != 0)
 		{
@@ -1233,23 +1244,23 @@ public class TextBox : TextBoxBase
 		ResetView();
 	}
 
-
 	// Paint the text using layout information
 	private void DrawText(Graphics g, bool focused)
 	{
 		Font font = Font;
-		StringBuilder s = new StringBuilder();
+		int lineStart = 0;
 		Brush prevFore = null;
 		int x = -1;
 		int y = -1;
 		int textLength = 0;
 		if (text != null)
 			textLength = text.Length;
-		for (int i=0; i < textLength;i++) 
+		int i;
+		for (i=0; i < textLength;i++) 
 		{
 			LayoutInfo.Item item = layout.Items[i];
 			Rectangle bounds = item.bounds;
-			bounds.Offset(-XViewOffset, -YViewOffset);
+			bounds.Offset(-xViewOffset, -yViewOffset);
 			if (item.type == LayoutInfo.Item.CharType.VisibleChar)
 			{
 				Brush fore;
@@ -1257,39 +1268,39 @@ public class TextBox : TextBoxBase
 					fore = SelectedForeBrush;
 				else
 					fore = ForeBrush;
-				if (bounds.Top == y && prevFore == fore)
+				if (bounds.Top != y || prevFore != fore)
 				{
-					if (passwordChar != 0)
-						s.Append(passwordChar);
-					else
-						s.Append(text[i]);
-				}
-				else
-				{
-					if (s.Length > 0)
+					if (i - lineStart > 0 && y < Height)
 					{
-						if (Enabled )
-							g.DrawString(s.ToString(), font, prevFore, new Point (x, y));
+						String lineText;
+						if (passwordChar != 0)
+							lineText = new String(passwordChar, i - lineStart);
 						else
-							ControlPaint.DrawStringDisabled(g, s.ToString(), font, BackColor, new Rectangle(x, y, int.MaxValue, int.MaxValue), StringFormat.GenericDefault);
+							lineText = text.Substring(lineStart, i - lineStart);
+						if (Enabled)
+							g.DrawString(lineText, font, prevFore, new Point (x, y));
+						else
+							ControlPaint.DrawStringDisabled(g, lineText, font, BackColor, new Rectangle(x, y, int.MaxValue, int.MaxValue), StringFormat.GenericDefault);
 					}
-					s.Remove(0, s.Length);
-					if (passwordChar != 0)
-						s.Append(passwordChar);
-					else
-						s.Append(text[i]);
+					lineStart = i;
 					x = bounds.X;
 					y = bounds.Y;
 				}
 				prevFore = fore;
 			}
 		}
-		if (s.Length != 0)
+		if (i - lineStart > 0)
 		{
-			if (Enabled )
-				g.DrawString(s.ToString(), Font, prevFore, new Point (x, y));
+			String lineText;
+			if (passwordChar != 0)
+				lineText = new String(passwordChar, i - lineStart);
 			else
-				ControlPaint.DrawStringDisabled(g, s.ToString(), font, BackColor, new Rectangle(x, y, int.MaxValue, int.MaxValue), StringFormat.GenericDefault);
+				lineText = text.Substring(lineStart, i - lineStart);
+						
+			if (Enabled )
+				g.DrawString(lineText, Font, prevFore, new Point (x, y));
+			else
+				ControlPaint.DrawStringDisabled(g, lineText, font, BackColor, new Rectangle(x, y, int.MaxValue, int.MaxValue), StringFormat.GenericDefault);
 		}
 	}
 
@@ -1404,8 +1415,10 @@ public class TextBox : TextBoxBase
 
 	// Set the caret bounds from a character position
 	// Set update region
-	private void CaretSetPosition( int position)
+	internal override void CaretSetPosition( int position)
 	{
+		if (!IsHandleCreated)
+			return;
 		Rectangle newBounds = Rectangle.Empty;
 		int height = Font.Height;
 		if (Text.Length == 0)
@@ -1533,13 +1546,6 @@ public class TextBox : TextBoxBase
 		}
 	}
 
-	// Sets the caret position to the end of the selection
-	private void CaretSetEndSelection()
-	{
-		// Even if we select backwards, the Caret will be at the end
-		CaretSetPosition(GetSelectionStart()+ GetSelectionLength());
-	}
-
 	// Change the selection and move the caret. Also make sure its visible
 	private void UpdateSelectionInternal(int newPos)
 	{
@@ -1693,7 +1699,7 @@ public class TextBox : TextBoxBase
 
 	private void HandleWordWrapChanged(object sender, EventArgs e)
 	{
-		SetupScrollBars();
+		layout = null;
 	}
 
 	private Rectangle TextDrawArea
