@@ -477,11 +477,11 @@ public class ToolBar : Control
 	                                                bool pushed)
 	{
 		ButtonState state = 0;
-		if (flat && !hover && !click)
+		if (flat && !hover && !click && !pushed)
 		{
 			state |= ButtonState.Flat;
 		}
-		if (click)
+		if (click || pushed)
 		{
 			state |= ButtonState.Pushed;
 		}
@@ -489,10 +489,6 @@ public class ToolBar : Control
 		{
 			state |= ButtonState.Inactive;
 		}
-		if (pushed)
-		{
-			state |= ButtonState.Checked;
-		}
 		return state;
 	}
 	private Size CalculateSizeH(int limitWidth)
@@ -730,6 +726,18 @@ public class ToolBar : Control
 #endif
 	protected override void Dispose(bool disposing)
 	{
+		if (disposing)
+		{
+			if (imageList != null)
+			{
+				imageList.RecreateHandle -= new EventHandler(ImageListHandler);
+#if CONFIG_COMPONENT_MODEL
+				imageList.Dispose();
+#else
+				imageList.Dispose(disposing);
+#endif
+			}
+		}
 		base.Dispose(disposing);
 	}
 	private void Draw(Graphics g)
