@@ -69,6 +69,7 @@ public class XmlTextReader : XmlReader
 	private String internalSubset;
 	private bool isEntity = false;
 	private int startEntity = 0;
+	private bool isDocType = false;
 
 	// Constructors.
 	protected XmlTextReader()
@@ -108,7 +109,7 @@ public class XmlTextReader : XmlReader
 				attr = new XmlAttributeToken(nameTable,null,null);
 				attributes = new XmlAttributeCollection(attr); 
 				attributeIndex = -1;
-				depth = 0;
+				depth = -1;
 				isEmpty = false;
 				contextSupport = false;
 				name = String.Empty;
@@ -360,7 +361,6 @@ public class XmlTextReader : XmlReader
 				StringReader s = new StringReader(tmp);
 				Close();
 				readState = ReadState.EndOfFile;
-				
 				return s;
 			}
 
@@ -547,6 +547,7 @@ public class XmlTextReader : XmlReader
 					else if (Char.IsWhiteSpace((char)ch))
 					{
 						SkipWhite();
+						ch = reader.Peek();
 						// more error checking
 						if((char)ch != '>' && (char)ch != '/')
 						{
@@ -807,6 +808,7 @@ public class XmlTextReader : XmlReader
 								if(builder.ToString(0, builder.Length -1) == "xml")
 								{
 									nodeType = XmlNodeType.XmlDeclaration;
+									SetName("xml");
 								}
 								else
 								{
@@ -823,11 +825,10 @@ public class XmlTextReader : XmlReader
 							if((char)ch == '?' && (char)reader.Peek() == '>')
 							{
 								value = builder.ToString(0, builder.Length -1);
-								ReadChar();
+								ch = ReadChar();
 								break;
 							}
 						}
-						
 						break;
 					case '!':
 						// Check for correct structure 
@@ -837,8 +838,9 @@ public class XmlTextReader : XmlReader
 						}
 
 						// Comment, CDATA, or document type information.
+						isDocType = true;
 						ch = ReadChar();
-						AnalyzeChar(ch, true);
+						AnalyzeChar(ch, false);
 						break;
 					case '-':
 						// Parse the "<!--" comment start sequence.
@@ -926,7 +928,7 @@ public class XmlTextReader : XmlReader
 						break;
 							
 					case 'D':
-						if(structFlag != true)
+						if(structFlag == true || isDocType == false)
 						{
 							goto default;
 						}
@@ -1757,7 +1759,7 @@ public class XmlTextReader : XmlReader
 			{
 				get
 				{
-					if(attributeIndex == -1)
+					if(attributeIndex == -1 && depth != -1)
 					{
 						return depth;
 					}
