@@ -38,28 +38,33 @@ public sealed class Display : IDisposable
 {
 	// Internal state.
 	internal IntPtr dpy;
-	private String displayName;
+	internal String displayName;
 	private Screen[] screens;
 	private int defaultScreen;
 	private Application app;
 	private bool quit;
 	private bool pendingExposes;
-	private bool inMainLoop;
 	private InputOutputWidget exposeList;
+	private InputOutputWidget invalidateList;
 	private Xlib.Cursor[] cursors;
 	internal Xlib.Time knownEventTime;
-	internal Hashtable handleMap;
+	internal HandleMap handleMap;
 	private static bool threadsInited;
 	internal Xlib.Atom wmProtocols;
 	internal Xlib.Atom wmDeleteWindow;
 	internal Xlib.Atom wmTakeFocus;
 	internal Xlib.Atom wmMwmHints;
+	internal Xlib.Atom wmContextHelp;
+	internal Xlib.Atom wmState;
+	internal Xlib.Atom wmNetState;
+	internal Xlib.Atom wmPing;
 	internal ButtonName selectButton;
 	internal ButtonName menuButton;
 	internal Hashtable fonts;
 	internal BuiltinBitmaps bitmaps;
 	internal Timer timerQueue;
 	internal IntPtr imlibData;
+	internal Xlib.Window groupLeader;
 
 	// Constructor.
 	private Display(IntPtr dpy, String displayName, Application app)
@@ -90,7 +95,7 @@ public sealed class Display : IDisposable
 				// Construct the window handle map if not already present.
 				if(handleMap == null)
 				{
-					handleMap = new Hashtable();
+					handleMap = new HandleMap();
 				}
 
 				// Initialize the standard window manager atoms that we use.
@@ -102,6 +107,14 @@ public sealed class Display : IDisposable
 					(dpy, "WM_TAKE_FOCUS", Xlib.Bool.False);
 				wmMwmHints = Xlib.XInternAtom
 					(dpy, "_MOTIF_WM_HINTS", Xlib.Bool.False);
+				wmContextHelp = Xlib.XInternAtom
+					(dpy, "_NET_WM_CONTEXT_HELP", Xlib.Bool.False);
+				wmState = Xlib.XInternAtom
+					(dpy, "WM_STATE", Xlib.Bool.False);
+				wmNetState = Xlib.XInternAtom
+					(dpy, "_NET_WM_STATE", Xlib.Bool.False);
+				wmPing = Xlib.XInternAtom
+					(dpy, "_NET_WM_PING", Xlib.Bool.False);
 
 				// Which buttons should we use for "Select" and "Menu"?
 				byte[] buttons = new byte [5];
@@ -145,10 +158,23 @@ public sealed class Display : IDisposable
 
 	// Internal version of "Open()" that is called once the
 	// type lock has been acquired.
-	private static Display OpenInternal(String displayName, Application app)
+	private static Display OpenInternal
+				(String displayName, Application app, bool sync)
 			{
 				try
 				{
+					if(Xlib.XSharpSupportPresent() == 0)
+					{
+						Console.WriteLine();
+						Console.WriteLine
+							("The XsharpSupport library appears to have " +
+							 "been built without X support.");
+						Console.WriteLine
+							("Please rebuild pnetlib to rectify this problem.");
+						Console.WriteLine();
+						throw new XCannotConnectException
+							(S._("X_LibraryNotPresent"));
+					}
 					try
 					{
 						// Initialize Xlib thread support.
@@ -172,6 +198,10 @@ public sealed class Display : IDisposable
 					if(dpy != IntPtr.Zero)
 					{
 						// We have opened the display successfully.
+						if(sync)
+						{
+							Xlib.XSynchronize(dpy, Xlib.Bool.True);
+						}
 						return new Display(dpy, displayName, app);
 					}
 					else
@@ -211,11 +241,11 @@ public sealed class Display : IDisposable
 			}
 
 	// Open a connection to a specific X display server.
-	internal static Display Open(String displayName, Application app)
+	internal static Display Open(String displayName, Application app, bool sync)
 			{
 				lock(typeof(Display))
 				{
-					return OpenInternal(displayName, app);
+					return OpenInternal(displayName, app, sync);
 				}
 			}
 
@@ -492,21 +522,12 @@ public sealed class Display : IDisposable
 	// Handle the next event and return what kind of event it was.
 	private AppEvent HandleNextEvent(bool wait)
 			{
-				bool isMainLoop = false;
 				try
 				{
 					IntPtr dpy = Lock();
 					XEvent xevent;
 					int timeout;
 
-					// Check for re-entry from multiple threads.
-					if(inMainLoop)
-					{
-						return AppEvent.NoEvent;
-					}
-					isMainLoop = true;
-					inMainLoop = true;
-
 					// Flush any requests that are in the outgoing queue.
 					Xlib.XFlush(dpy);
 	
@@ -515,6 +536,9 @@ public sealed class Display : IDisposable
 					{
 						return AppEvent.Quit;
 					}
+
+					// Processing any pending invalidates that we have.
+					ProcessPendingInvalidates();
 	
 					// Do we have pending expose events to process?
 					if(pendingExposes)
@@ -615,10 +639,6 @@ public sealed class Display : IDisposable
 				}
 				finally
 				{
-					if(isMainLoop)
-					{
-						inMainLoop = false;
-					}
 					Unlock();
 				}
 
@@ -681,14 +701,27 @@ public sealed class Display : IDisposable
 	// have the display lock.
 	private void DispatchEvent(ref XEvent xevent)
 			{
+				// Find the widget that should process the event.
+				Widget widget = handleMap[xevent.xany.window];
+
 				// Record the time at which the event occurred.  We need
 				// this to process keyboard and pointer grabs correctly.
-				switch(xevent.type)
+				switch((EventType)(xevent.xany.type__))
 				{
 					case EventType.KeyPress:
 					case EventType.KeyRelease:
 					{
 						knownEventTime = xevent.xkey.time;
+						if(widget != null && !(widget.Parent is RootWindow))
+						{
+							// KeyPress/KeyRelease events must be dispatched
+							// via the top-level window, never via children.
+							while(widget.Parent != null &&
+							      !(widget.Parent is RootWindow))
+							{
+								widget = widget.Parent;
+							}
+						}
 					}
 					break;
 
@@ -744,9 +777,6 @@ public sealed class Display : IDisposable
 					break;
 				}
 
-				// Find the widget that should process the event.
-				Widget widget = (Widget)(handleMap[(int)(xevent.window)]);
-
 				// Dispatch the event to the widget.
 				if(widget != null)
 				{
@@ -812,6 +842,51 @@ public sealed class Display : IDisposable
 				}
 			}
 
+	// Add an input/output widget to the pending invalidate list.
+	internal void AddPendingInvalidate(InputOutputWidget widget)
+			{
+				widget.nextInvalidate = invalidateList;
+				invalidateList = widget;
+			}
+
+	// Remove an input/output widget from the pending invalidate list.
+	internal void RemovePendingInvalidate(InputOutputWidget widget)
+			{
+				InputOutputWidget current, prev;
+				current = invalidateList;
+				prev = null;
+				while(current != null && current != widget)
+				{
+					prev = current;
+					current = current.nextInvalidate;
+				}
+				if(current != null)
+				{
+					if(prev != null)
+					{
+						prev.nextInvalidate = current.nextInvalidate;
+					}
+					else
+					{
+						invalidateList = current.nextInvalidate;
+					}
+				}
+			}
+
+	// Process pending invalidates.
+	private void ProcessPendingInvalidates()
+			{
+				InputOutputWidget current = invalidateList;
+				InputOutputWidget next;
+				invalidateList = null;
+				while(current != null)
+				{
+					next = current.nextInvalidate;
+					current.FlushInvalidates();
+					current = next;
+				}
+			}
+
 	// Get the maximum request size for the display.
 	internal long MaxRequestSize()
 			{
