@@ -56,10 +56,10 @@ using Xsharp.Types;
 public sealed class Graphics : IDisposable
 {
 	// Internal state.
-	private Display dpy;
+	internal Display dpy;
 	private Drawable drawable;
-	private Xlib.Drawable drawableHandle;
-	private IntPtr gc;
+	internal Xlib.Drawable drawableHandle;
+	internal IntPtr gc;
 	private Color foreground;
 	private Color background;
 	private Pixmap tile;
@@ -363,6 +363,56 @@ public sealed class Graphics : IDisposable
 				}
 			}
 
+	/// <summary>
+	/// <para>Get or set the subwindow mode.</para>
+	/// </summary>
+	///
+	/// <value>
+	/// <para>The <see cref="T:Xsharp.SubwindowMode"/> value
+	/// for the mode.</para>
+	/// </value>
+	///
+	/// <exception cref="T:Xsharp.XException">
+	/// <para>Raised if set to an invalid value.</para>
+	/// </exception>
+	public Xsharp.SubwindowMode SubwindowMode
+			{
+				get
+				{
+					try
+					{
+						IntPtr display = Lock();
+						XGCValues values;
+						Xlib.XGetGCValues(display, gc,
+										  (uint)(GCValueMask.GCSubwindowMode),
+										  out values);
+						return values.subwindow_mode;
+					}
+					finally
+					{
+						dpy.Unlock();
+					}
+				}
+				set
+				{
+					if(value < Xsharp.SubwindowMode.ClipByChildren ||
+					   value > Xsharp.SubwindowMode.IncludeInferiors)
+					{
+						throw new XException
+							(String.Format(S._("X_SubwindowMode"), (int)value));
+					}
+					try
+					{
+						IntPtr display = Lock();
+						Xlib.XSetSubwindowMode(display, gc, (int)value);
+					}
+					finally
+					{
+						dpy.Unlock();
+					}
+				}
+			}
+
 	/// <summary>
 	/// <para>Get or set the line width.</para>
 	/// </summary>
@@ -1152,7 +1202,7 @@ public sealed class Graphics : IDisposable
 				try
 				{
 					IntPtr display = Lock();
-					Xlib.XSetClipOrigin(display, gc, xorigin, xorigin);
+					Xlib.XSetClipOrigin(display, gc, xorigin, yorigin);
 					if(mask != null)
 					{
 						Xlib.XSetClipMask(display, gc, mask.GetPixmapHandle());
@@ -1653,8 +1703,8 @@ public sealed class Graphics : IDisposable
 	/// </exception>
 	public void DrawRectangle(int x, int y, int width, int height)
 			{
-				if(x < -32768 || x > 32767 || width < 0 || width > 65535 ||
-				   y < -32768 || y > 32767 || height < 0 || height > 65535)
+				if(x < -32768 || x > 32767 || width <  -32768 || width > 32767 ||
+				   y < -32768 || y > 32767 || height < -32768 || height > 32767)
 				{
 					throw new XException(S._("X_RectCoordRange"));
 				}
@@ -1720,8 +1770,8 @@ public sealed class Graphics : IDisposable
 	/// </exception>
 	public void FillRectangle(int x, int y, int width, int height)
 			{
-				if(x < -32768 || x > 32767 || width < 0 || width > 65535 ||
-				   y < -32768 || y > 32767 || height < 0 || height > 65535)
+				if(x < -32768 || x > 32767 || width < -32768 || width > 32767 ||
+				   y < -32768 || y > 32767 || height < -32768 || height > 32767)
 				{
 					throw new XException(S._("X_RectCoordRange"));
 				}
@@ -1912,8 +1962,8 @@ public sealed class Graphics : IDisposable
 	public void DrawArc(int x, int y, int width, int height,
 						float startAngle, float sweepAngle)
 			{
-				if(x < -32768 || x > 32767 || width < 0 || width > 65535 ||
-				   y < -32768 || y > 32767 || height < 0 || height > 65535)
+				if(x < -32768 || x > 32767 || width < -32768 || width > 32767 ||
+				   y < -32768 || y > 32767 || height < -32768 || height > 32767)
 				{
 					throw new XException(S._("X_RectCoordRange"));
 				}
@@ -1972,8 +2022,8 @@ public sealed class Graphics : IDisposable
 	public void DrawPie(int x, int y, int width, int height,
 						float startAngle, float sweepAngle)
 			{
-				if(x < -32768 || x > 32767 || width < 0 || width > 65535 ||
-				   y < -32768 || y > 32767 || height < 0 || height > 65535)
+				if(x < -32768 || x > 32767 || width < -32768 || width > 32767 ||
+				   y < -32768 || y > 32767 || height < -32768 || height > 32767)
 				{
 					throw new XException(S._("X_RectCoordRange"));
 				}
@@ -2057,8 +2107,8 @@ public sealed class Graphics : IDisposable
 	public void FillArc(int x, int y, int width, int height,
 						float startAngle, float sweepAngle)
 			{
-				if(x < -32768 || x > 32767 || width < 0 || width > 65535 ||
-				   y < -32768 || y > 32767 || height < 0 || height > 65535)
+				if(x < -32768 || x > 32767 || width < -32768 || width > 32767 ||
+				   y < -32768 || y > 32767 || height < -32768 || height > 32767)
 				{
 					throw new XException(S._("X_RectCoordRange"));
 				}
@@ -2263,11 +2313,6 @@ public sealed class Graphics : IDisposable
 	/// </exception>
 	public void DrawString(int x, int y, String str, Font font)
 			{
-				// Validate the parameters.
-				if(x < -32768 || x > 32767 || y < -32768 || y > 32767)
-				{
-					throw new XException(S._("X_PointCoordRange"));
-				}
 				if(font == null)
 				{
 					throw new ArgumentNullException("font");
@@ -2276,27 +2321,7 @@ public sealed class Graphics : IDisposable
 				{
 					return;
 				}
-
-				// Get the font set to use for the font.
-				IntPtr fontSet = font.GetFontSet(dpy);
-				if(fontSet == IntPtr.Zero)
-				{
-					return;
-				}
-
-				// Draw the string using the specified font set.
-				try
-				{
-					IntPtr display = dpy.Lock();
-					Xlib.XSharpDrawString
-							(display, drawableHandle, gc,
-							 fontSet, x, y, str, (int)(font.Style),
-							 IntPtr.Zero, foreground.value);
-				}
-				finally
-				{
-					dpy.Unlock();
-				}
+				font.DrawString(this, x, y, str, 0, str.Length);
 			}
 
 	/// <summary>
@@ -2322,9 +2347,7 @@ public sealed class Graphics : IDisposable
 				{
 					throw new ArgumentNullException("font");
 				}
-				FontExtents extents = null;
-				font.GetFontSet(dpy, out extents);
-				return extents;
+				return font.GetFontExtents(this);
 			}
 
 	/// <summary>
@@ -2360,7 +2383,6 @@ public sealed class Graphics : IDisposable
 	public void MeasureString(String str, Font font, out int width,
 							  out int ascent, out int descent)
 			{
-				// Validate the parameters.
 				if(font == null)
 				{
 					throw new ArgumentNullException("font");
@@ -2372,45 +2394,8 @@ public sealed class Graphics : IDisposable
 					descent = 0;
 					return;
 				}
-
-				// Get the font set to use to measure the string.
-				IntPtr fontSet = font.GetFontSet(dpy);
-				if(fontSet == IntPtr.Zero)
-				{
-					width = 0;
-					ascent = 0;
-					descent = 0;
-					return;
-				}
-
-				// Get the text extents and decode them into useful values.
-				XRectangle overall_ink;
-				XRectangle overall_logical;
-				try
-				{
-					IntPtr display = dpy.Lock();
-					Xlib.XSharpTextExtents
-						(display, fontSet, str,
-						 out overall_ink, out overall_logical);
-				}
-				finally
-				{
-					dpy.Unlock();
-				}
-				width = overall_logical.width;
-				ascent = -(overall_logical.y);
-				descent = overall_logical.height + overall_logical.y;
-
-				// Increase the descent to account for underlining.
-				// We always draw the underline two pixels below
-				// the font base line.
-				if((font.Style & FontStyle.Underlined) != 0)
-				{
-					if(descent < 3)
-					{
-						descent = 3;
-					}
-				}
+				font.MeasureString(this, str, 0, str.Length,
+								   out width, out ascent, out descent);
 			}
 
 	// Draw a radio button.
@@ -2470,6 +2455,32 @@ public sealed class Graphics : IDisposable
 				}
 			}
 
+	// Draw a simple bitmap.
+	internal void DrawBitmap(int x, int y, int width, int height,
+							 Xlib.Pixmap bitmap)
+			{
+				try
+				{
+					IntPtr display = Lock();
+					Xlib.XSetStipple(display, gc, bitmap);
+					Xlib.XSetTSOrigin(display, gc, x, y);
+					Xlib.XSetForeground(display, gc,
+						drawable.ToPixel
+							(new Color(StandardColor.Foreground)));
+					Xlib.XSetFillStyle
+						(display, gc, (int)(FillStyle.FillStippled));
+					Xlib.XFillRectangle(display, drawableHandle, gc,
+										x, y, width, height);
+					Xlib.XSetTSOrigin(display, gc, 0, 0);
+					Xlib.XSetFillStyle
+						(display, gc, (int)(FillStyle.FillSolid));
+				}
+				finally
+				{
+					dpy.Unlock();
+				}
+			}
+
 	/// <summary>
 	/// <para>Draw a three-dimensional effect.</para>
 	/// </summary>
@@ -2516,6 +2527,7 @@ public sealed class Graphics : IDisposable
 					Xlib.Pixel topShadowEnhance;
 					Xlib.Pixel bottomShadow;
 					Xlib.Pixel bottomShadowEnhance;
+					Xlib.Pixel background;
 					Xlib.Pixel trim;
 					if((effect & Effect.ContentColors) != 0)
 					{
@@ -2527,6 +2539,8 @@ public sealed class Graphics : IDisposable
 							(new Color(StandardColor.ContentBottomShadow));
 						bottomShadowEnhance = drawable.ToPixel(new Color
 							(StandardColor.ContentBottomShadowEnhance));
+						background = drawable.ToPixel
+							(new Color(StandardColor.ContentBackground));
 						trim = drawable.ToPixel
 							(new Color(StandardColor.ContentTrim));
 					}
@@ -2540,6 +2554,8 @@ public sealed class Graphics : IDisposable
 							(new Color(StandardColor.BottomShadow));
 						bottomShadowEnhance = drawable.ToPixel
 							(new Color(StandardColor.BottomShadowEnhance));
+						background = drawable.ToPixel
+							(new Color(StandardColor.Background));
 						trim = drawable.ToPixel(new Color(StandardColor.Trim));
 					}
 
@@ -2563,12 +2579,14 @@ public sealed class Graphics : IDisposable
 					newValues.line_style = LineStyle.LineSolid;
 					newValues.join_style = JoinStyle.JoinMiter;
 					newValues.cap_style = CapStyle.CapProjecting;
+					newValues.fill_style = FillStyle.FillSolid;
 					Xlib.XChangeGC(display, gc,
 								   (uint)(GCValueMask.GCFunction |
 								  		  GCValueMask.GCLineWidth |
 										  GCValueMask.GCLineStyle |
 										  GCValueMask.GCJoinStyle |
-										  GCValueMask.GCCapStyle),
+										  GCValueMask.GCCapStyle |
+										  GCValueMask.GCFillStyle),
 								   ref newValues);
 
 					// Draw the effect.
@@ -2802,6 +2820,68 @@ public sealed class Graphics : IDisposable
 							// TODO
 						}
 						break;
+
+						case Effect.CaptionButtonRaised:
+						{
+							if(width >= 4 && height >= 4)
+							{
+								Xlib.XSetForeground(display, gc, topShadow);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y, x + width - 1, y);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y + 1, x, y + height - 2);
+								Xlib.XSetForeground(display, gc, bottomShadow);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + 1, y + height - 2, x + width - 2,
+									y + height - 2);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + width - 2, y + 1, x + width - 2,
+									y + height - 3);
+								Xlib.XSetForeground
+									(display, gc, bottomShadowEnhance);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y + height - 1, x + width - 1,
+									y + height - 1);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + width - 1, y, x + width - 1,
+									y + height - 2);
+								Xlib.XSetForeground
+									(display, gc, background);
+								Xlib.XFillRectangle(display, drawableHandle, gc,
+									x + 1, y + 1, width - 3, height - 3);
+							}
+						}
+						break;
+
+						case Effect.CaptionButtonIndented:
+						{
+							if(width >= 4 && height >= 4)
+							{
+								Xlib.XSetForeground(display, gc, bottomShadow);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y, x + width - 2, y);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y + 1, x, y + height - 2);
+								Xlib.XSetForeground
+									(display, gc, bottomShadowEnhance);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + 1, y + 1, x + width - 3, y + 1);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + 1, y + 2, x + 1, y + height - 3);
+								Xlib.XSetForeground(display, gc, topShadow);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x, y + height - 1, x + width - 1,
+									y + height - 1);
+								Xlib.XDrawLine(display, drawableHandle, gc,
+									x + width - 1, y, x + width - 1,
+									y + height - 2);
+								Xlib.XSetForeground
+									(display, gc, background);
+								Xlib.XFillRectangle(display, drawableHandle, gc,
+									x + 2, y + 2, width - 3, height - 3);
+							}
+						}
+						break;
 					}
 
 					// Restore the previous GC settings.
