@@ -35,8 +35,8 @@ public class InputOnlyWidget : Widget
 {
 	// Internal state.
 	private bool focusable;
-	private Xlib.Time lastClickTime;
-	private ButtonName lastClickButton;
+	internal Xlib.Time lastClickTime;
+	internal ButtonName lastClickButton;
 
 	/// <summary>
 	/// <para>Constructs a new <see cref="T:Xsharp.InputOnlyWidget"/>
@@ -115,13 +115,14 @@ public class InputOnlyWidget : Widget
 						Xlib.Window window = Xlib.XCreateWindow
 								(display, pwindow,
 								 x, y, (uint)width, (uint)height, (uint)0,
-								 screen.DefaultDepth, 2 /* InputOnly */,
+								 0 /* depth */, 2 /* InputOnly */,
 								 screen.DefaultVisual,
 								 (uint)0, IntPtr.Zero);
 						SetWidgetHandle(window);
 						if(parent.AutoMapChildren)
 						{
 							Xlib.XMapWindow(display, window);
+							mapped = true;
 						}
 					}
 					finally
@@ -147,7 +148,7 @@ public class InputOnlyWidget : Widget
 					{
 						// Creation failed, so detach ourselves from
 						// the parent's widget tree.
-						Detach();
+						Detach(false);
 					}
 				}
 			}
@@ -210,6 +211,7 @@ public class InputOnlyWidget : Widget
 						if(parent.AutoMapChildren)
 						{
 							Xlib.XMapWindow(display, window);
+							mapped = true;
 						}
 					}
 					finally
@@ -235,7 +237,7 @@ public class InputOnlyWidget : Widget
 					{
 						// Creation failed, so detach ourselves from
 						// the parent's widget tree.
-						Detach();
+						Detach(false);
 					}
 				}
 			}
@@ -619,7 +621,7 @@ public class InputOnlyWidget : Widget
 				ButtonName button;
 				Xlib.Time time;
 	
-				switch(xevent.type)
+				switch((EventType)(xevent.xany.type__))
 				{
 					case EventType.ButtonPress:
 					{
@@ -668,8 +670,8 @@ public class InputOnlyWidget : Widget
 					case EventType.EnterNotify:
 					{
 						// Dispatch a widget enter event.
-						Widget child = (Widget)(dpy.handleMap
-							[(int)(xevent.xcrossing.subwindow)]);
+						Widget child = dpy.handleMap
+							[xevent.xcrossing.subwindow];
 						OnEnter(child,
 							    xevent.xcrossing.x,
 							    xevent.xcrossing.y,
@@ -682,8 +684,8 @@ public class InputOnlyWidget : Widget
 					case EventType.LeaveNotify:
 					{
 						// Dispatch a widget leave event.
-						Widget child = (Widget)(dpy.handleMap
-							[(int)(xevent.xcrossing.subwindow)]);
+						Widget child = dpy.handleMap
+							[xevent.xcrossing.subwindow];
 						OnLeave(child,
 							    xevent.xcrossing.x,
 							    xevent.xcrossing.y,
