@@ -44,7 +44,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 	internal bool autoMapChildren;
 	private bool sensitive;
 	private bool ancestorSensitive;
-	private CursorType cursor;
+	private Cursor cursor;
 	private Widget parent;
 	private Widget topChild;
 	private Widget nextAbove;
@@ -57,7 +57,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 			: base(dpy, screen, kind)
 			{
 				// Set the initial widget properties.
-				cursor = CursorType.XC_inherit_parent;
+				cursor = null;
 				autoMapChildren = true;
 				sensitive = true;
 
@@ -85,7 +85,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 			}
 
 	// Detach this widget from its position in the widget tree.
-	internal void Detach()
+	internal void Detach(bool parentOnly)
 			{
 				// Detach ourselves from our siblings and parent.
 				if(nextBelow != null)
@@ -102,20 +102,23 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 				}
 
 				// Detach ourselves from our children.
-				Widget current, next;
-				current = topChild;
-				while(current != null)
+				if(!parentOnly)
 				{
-					next = current.nextBelow;
-					current.parent = null;
-					current.nextAbove = null;
-					current.nextBelow = null;
-					current = next;
+					Widget current, next;
+					current = topChild;
+					while(current != null)
+					{
+						next = current.nextBelow;
+						current.parent = null;
+						current.nextAbove = null;
+						current.nextBelow = null;
+						current = next;
+					}
+					topChild = null;
 				}
 
 				// Clear all of our link fields.
 				parent = null;
-				topChild = null;
 				nextAbove = null;
 				nextBelow = null;
 			}
@@ -123,11 +126,11 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 	// Disassociate this widget instance and all of its children
 	// from their X window handles, as the mirror copy in the X
 	// server has been lost.
-	internal void Disassociate()
+	internal virtual void Disassociate()
 			{
 				if(handle != Xlib.Drawable.Zero)
 				{
-					dpy.handleMap.Remove((int)handle);
+					dpy.handleMap.Remove((Xlib.Window)handle);
 				}
 				if(this is InputOutputWidget)
 				{
@@ -152,8 +155,9 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 					IntPtr d = dpy.Lock();
 					if(handle != Xlib.Drawable.Zero)
 					{
-						Xlib.XDestroyWindow(d, (Xlib.Window)handle);
+						Xlib.Drawable tempHandle = handle;
 						Disassociate();
+						Xlib.XDestroyWindow(d, (Xlib.Window)tempHandle);
 					}
 				}
 				finally
@@ -162,6 +166,23 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 				}
 			}
 
+	/// <summary>
+	/// <para>Destroy this drawable from a finalizer if it is
+	/// currently active.</para>
+	/// </summary>
+	///
+	/// <remarks>
+	/// <para>This version is called from object finalizers.  Subclasses
+	/// need to be careful not to lock the display during finalization
+	/// because doing so can lead to deadlock conditions.</para>
+	/// </remarks>
+	public override void DestroyFromFinalizer()
+			{
+				// Just clear the handle: the X server will clean
+				// up the window object for us at shutdown.
+				handle = Xlib.Drawable.Zero;
+			}
+
 	/// <summary>
 	/// <para>Get the X position of this widget relative to its parent.</para>
 	/// </summary>
@@ -260,10 +281,10 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 	///
 	/// <value>
 	/// <para>The cursor shape to set for the widget.  If the value is
-	/// <c>CursorType.XC_inherit_parent</c>, then the widget inherits the
+	/// <see langword="null"/>, then the widget inherits the
 	/// cursor that is set on the parent widget.</para>
 	/// </value>
-	public virtual CursorType Cursor
+	public virtual Cursor Cursor
 			{
 				get
 				{
@@ -274,10 +295,10 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 					try
 					{
 						IntPtr display = dpy.Lock();
-						if(cursor != value)
+						if(handle != Xlib.Drawable.Zero && cursor != value)
 						{
 							cursor = value;
-							if(value == CursorType.XC_inherit_parent)
+							if(value == null)
 							{
 								// Revert to inheriting our parent's cursor.
 								Xlib.XUndefineCursor
@@ -285,10 +306,8 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 							}
 							else
 							{
-								// Change our cursor to a pre-defined shape.
-								Xlib.XDefineCursor
-									(display, GetWidgetHandle(),
-									 dpy.GetCursor(value));
+								// Change our cursor to the defined shape.
+								value.SetCursor(this);
 							}
 						}
 					}
@@ -850,7 +869,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 					{
 						AdjustPositionAndSize(display, x, y,
 											  this.width, this.height);
-						OnMove(x, y);
+						OnMoveResize(x, y, this.width, this.height);
 					}
 				}
 				finally
@@ -889,7 +908,59 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 					{
 						AdjustPositionAndSize(display, this.x, this.y,
 											  width, height);
-						OnResize(width, height);
+						OnMoveResize(this.x, this.y, width, height);
+					}
+				}
+				finally
+				{
+					dpy.Unlock();
+				}
+			}
+
+	/// <summary>
+	/// <para>Move and resize this widget.</para>
+	/// </summary>
+	///
+	/// <param name="x">
+	/// <para>The X co-ordinate of the new top-left widget corner.</para>
+	/// </param>
+	///
+	/// <param name="y">
+	/// <para>The Y co-ordinate of the new top-left widget corner.</para>
+	/// </param>
+	///
+	/// <param name="width">
+	/// <para>The new width for the widget.</para>
+	/// </param>
+	///
+	/// <param name="height">
+	/// <para>The new width for the widget.</para>
+	/// </param>
+	///
+	/// <exception cref="T:Xsharp.XException">
+	/// <para>Raised if <paramref name="width"/> or <paramref name="height"/>
+	/// is out of range.</para>
+	/// </exception>
+	public virtual void MoveResize(int x, int y, int width, int height)
+			{
+				if(width < 1 || height < 1 ||
+				   !ValidateSize(width, height))
+				{
+					throw new XException(S._("X_InvalidSize"));
+				}
+				try
+				{
+					IntPtr display = dpy.Lock();
+					bool moved = (x != this.x || y != this.y);
+					bool resized = (width != this.width ||
+									height != this.height);
+					if(moved || resized)
+					{
+						AdjustPositionAndSize(display, x, y, width, height);
+						if(moved || resized)
+						{
+							OnMoveResize(x, y, width, height);
+						}
 					}
 				}
 				finally
@@ -1404,7 +1475,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 
 	/// <summary>
 	/// <para>Method that is called when the widget is moved to a
-	/// new position.</para>
+	/// new position or given a new size.</para>
 	/// </summary>
 	///
 	/// <param name="x">
@@ -1414,15 +1485,6 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 	/// <param name="y">
 	/// <para>The Y co-ordinate of the new top-left widget corner.</para>
 	/// </param>
-	protected virtual void OnMove(int x, int y)
-			{
-				// Nothing to do in the base class.
-			}
-
-	/// <summary>
-	/// <para>Method that is called when the widget is resized to a
-	/// new size.</para>
-	/// </summary>
 	///
 	/// <param name="width">
 	/// <para>The new width for the widget.</para>
@@ -1431,7 +1493,7 @@ public abstract class Widget : Drawable, ICollection, IEnumerable
 	/// <param name="height">
 	/// <para>The new width for the widget.</para>
 	/// </param>
-	protected virtual void OnResize(int width, int height)
+	protected virtual void OnMoveResize(int x, int y, int width, int height)
 			{
 				// Nothing to do in the base class.
 			}
