@@ -20,36 +20,12 @@
 
 #if !defined(X_DISPLAY_MISSING) && HAVE_SELECT
 
-#include <X11/Xlib.h>
-#include <X11/Xutil.h>
-#include <X11/Xatom.h>
-#ifdef WIN32
-	#include <X11/Xwinsock.h>
-#endif
-#ifdef USE_XFT_EXTENSION
-	#include <X11/Xft/Xft.h>
-#endif
-#if TIME_WITH_SYS_TIME
-	#include <sys/time.h>
-    #include <time.h>
-#else
-    #if HAVE_SYS_TIME_H
-		#include <sys/time.h>
-    #else
-        #include <time.h>
-    #endif
-#endif
-#if HAVE_SYS_TYPES_H
-	#include <sys/types.h>
-#endif
-#if HAVE_SYS_SELECT_H
-	#include <sys/select.h>
-#endif
-#if HAVE_UNISTD_H
-	#include <unistd.h>
-#endif
-#include <stdlib.h>
-#include <stdio.h>
+#include "XsharpSupport.h"
+
+int XSharpSupportPresent(void)
+{
+	return 1;
+}
 
 int XNextEventWithTimeout(Display *dpy, XEvent *event, int timeout)
 {
@@ -86,21 +62,22 @@ int XNextEventWithTimeout(Display *dpy, XEvent *event, int timeout)
 }
 
 /*
- * Font style flags.
+ * Determine if we can use the Xft extension.
  */
-#define	FontStyle_Normal		0
-#define	FontStyle_Bold			1
-#define	FontStyle_Italic		2
-#define	FontStyle_Underline		4
-#define	FontStyle_StrikeOut		8
-
-#ifndef USE_XFT_EXTENSION
+int XSharpUseXft()
+{
+#ifdef USE_XFT_EXTENSION
+	return 1;
+#else
+	return 0;
+#endif
+}
 
 /*
  * Try to create a font with specific parameters.
  */
-static XFontSet TryCreateFont(Display *dpy, const char *family,
-							  int pointSize, int style)
+static void *TryCreateFont(Display *dpy, const char *family,
+						   int pointSize, int style)
 {
 	XFontSet fontSet;
 	char *name;
@@ -147,6 +124,23 @@ static XFontSet TryCreateFont(Display *dpy, const char *family,
 				(family ? family : "*"), weight, slant);
 	}
 
+	/* Use the Latin1 XFontStruct creation method if requested */
+	if((style & FontStyle_FontStruct) != 0)
+	{
+		XFontStruct *fs;
+		fs = XLoadQueryFont(dpy, name);
+		if(fs)
+		{
+			free(name);
+			return (void *)fs;
+		}
+		else
+		{
+			free(name);
+			return 0;
+		}
+	}
+
 	/* Try to create the font set using just the base name */
 	missingCharsets = 0;
 	missingCharsetCount = 0;
@@ -156,7 +150,7 @@ static XFontSet TryCreateFont(Display *dpy, const char *family,
 	if(fontSet)
 	{
 		free(name);
-		return fontSet;
+		return (void *)fontSet;
 	}
 	/* TODO: process the missing charsets */
 
@@ -165,13 +159,63 @@ static XFontSet TryCreateFont(Display *dpy, const char *family,
 	return 0;
 }
 
-#endif /* USE_XFT_EXTENSION */
+/*
+ * Create a font set from a description.
+ */
+void *XSharpCreateFontSet(Display *dpy, const char *family,
+					      int pointSize, int style)
+{
+	XFontSet fontSet;
+	int structStyle = (style & FontStyle_FontStruct);
+
+	/* Try with the actual parameters first */
+	fontSet = TryCreateFont(dpy, family, pointSize, style);
+	if(fontSet)
+	{
+		return fontSet;
+	}
+
+	/* Remove the style, but keep the family */
+	fontSet = TryCreateFont(dpy, family, pointSize,
+							FontStyle_Normal | structStyle);
+	if(fontSet)
+	{
+		return fontSet;
+	}
+
+	/* Remove the family, but keep the style */
+	fontSet = TryCreateFont(dpy, 0, pointSize, style);
+	if(fontSet)
+	{
+		return fontSet;
+	}
+
+	/* Remove the point size and the style, but keep the family */
+	fontSet = TryCreateFont(dpy, family, -1, FontStyle_Normal | structStyle);
+	if(fontSet)
+	{
+		return fontSet;
+	}
+
+	/* Remove everything - this will succeed unless X has no fonts at all! */
+	return TryCreateFont(dpy, 0, -1, FontStyle_Normal | structStyle);
+}
 
 /*
- * Create a font from a description.
+ * Create a font struct from a description.
  */
-void *XSharpCreateFont(Display *dpy, const char *family,
-					   int pointSize, int style)
+void *XSharpCreateFontStruct(Display *dpy, const char *family,
+					         int pointSize, int style)
+{
+	return XSharpCreateFontSet(dpy, family, pointSize,
+							   style | FontStyle_FontStruct);
+}
+
+/*
+ * Create an Xft font from a description.
+ */
+void *XSharpCreateFontXft(Display *dpy, const char *family,
+					      int pointSize, int style)
 {
 #ifdef USE_XFT_EXTENSION
 
@@ -226,82 +270,217 @@ void *XSharpCreateFont(Display *dpy, const char *family,
 
 #else /* !USE_XFT_EXTENSION */
 
-	XFontSet fontSet;
+	/* Don't have Xft support on this platform */
+	return 0;
 
-	/* Try with the actual parameters first */
-	fontSet = TryCreateFont(dpy, family, pointSize, style);
-	if(fontSet)
-	{
-		return fontSet;
-	}
-
-	/* Remove the style, but keep the family */
-	fontSet = TryCreateFont(dpy, family, pointSize, FontStyle_Normal);
-	if(fontSet)
-	{
-		return fontSet;
-	}
-
-	/* Remove the family, but keep the style */
-	fontSet = TryCreateFont(dpy, 0, pointSize, style);
-	if(fontSet)
-	{
-		return fontSet;
-	}
-
-	/* Remove the point size and the style, but keep the family */
-	fontSet = TryCreateFont(dpy, family, -1, FontStyle_Normal);
-	if(fontSet)
-	{
-		return fontSet;
-	}
-
-	/* Remove everything - this will succeed unless X has no fonts at all! */
-	return TryCreateFont(dpy, 0, -1, FontStyle_Normal);
-
-#endif
+#endif /* !USE_XFT_EXTENSION */
 }
 
 /*
  * Free a font set.
  */
+void XSharpFreeFontSet(Display *dpy, void *fontSet)
+{
+	XFreeFontSet(dpy, (XFontSet)fontSet);
+}
+
+/*
+ * Free a font struct.
+ */
+void XSharpFreeFontStruct(Display *dpy, void *fontSet)
+{
+	XFreeFont(dpy, (XFontStruct *)fontSet);
+}
+
+/*
+ * Free an Xft font.
+ */
 void XSharpFreeFont(Display *dpy, void *fontSet)
 {
 #ifdef USE_XFT_EXTENSION
 	XftFontClose(dpy, (XftFont *)fontSet);
-#else
-	XFreeFontSet(dpy, (XFontSet)fontSet);
 #endif
 }
 
 /*
- * Forward declaration.
+ * Forward declarations.
  */
-void XSharpTextExtents(Display *dpy, void *fontSet, const char *str,
-					   XRectangle *overall_ink_return,
-					   XRectangle *overall_logical_return);
+void XSharpTextExtentsSet(Display *dpy, void *fontSet, const char *str,
+					      XRectangle *overall_ink_return,
+					      XRectangle *overall_logical_return);
+void XSharpTextExtentsStruct(Display *dpy, void *fontSet,
+					         ILString *str, long offset, long count,
+							 XRectangle *overall_ink_return,
+					         XRectangle *overall_logical_return);
+void XSharpTextExtentsXft(Display *dpy, void *fontSet, const char *str,
+					      XRectangle *overall_ink_return,
+					      XRectangle *overall_logical_return);
 
 /*
  * Draw a string using a font set.
  */
-void XSharpDrawString(Display *dpy, Drawable drawable, GC gc,
-					  void *fontSet, int x, int y,
-					  const char *str, int style, Region clipRegion,
-					  unsigned long colorValue)
+void XSharpDrawStringSet(Display *dpy, Drawable drawable, GC gc,
+					     void *fontSet, int x, int y,
+					     const char *str, int style)
 {
 	XRectangle overall_ink_return;
 	XRectangle overall_logical_return;
+	XFontSetExtents *extents;
+	int line1, line2;
+
+	/* Draw the string using the core API */
+	XmbDrawString(dpy, drawable, (XFontSet)fontSet, gc, x, y,
+				  str, strlen(str));
+
+	/* Calculate the positions of the underline and strike-out */
+	if((style & FontStyle_Underline) != 0)
+	{
+		line1 = y + 1;
+	}
+	else
+	{
+		line1 = y;
+	}
+	if((style & FontStyle_StrikeOut) != 0)
+	{
+		extents = XExtentsOfFontSet((XFontSet)fontSet);
+		if(extents)
+		{
+			line2 = y + (extents->max_logical_extent.y / 2);
+		}
+		else
+		{
+			line2 = y;
+		}
+	}
+	else
+	{
+		line2 = y;
+	}
+
+	/* Draw the underline and strike-out */
+	if(line1 != y || line2 != y)
+	{
+		XSharpTextExtentsSet(dpy, fontSet, str,
+				 	         &overall_ink_return, &overall_logical_return);
+		XSetLineAttributes(dpy, gc, 1, LineSolid, CapNotLast, JoinMiter);
+		if(line1 != y)
+		{
+			XDrawLine(dpy, drawable, gc, x, line1,
+					  x + overall_logical_return.width, line1);
+		}
+		if(line2 != y)
+		{
+			XDrawLine(dpy, drawable, gc, x, line2,
+					  x + overall_logical_return.width, line2);
+		}
+	}
+}
+
+#define	XSHARP_BUFSIZ	128
+
+/*
+ * Draw a string using a font struct.
+ */
+void XSharpDrawStringStruct(Display *dpy, Drawable drawable, GC gc,
+					        void *fontSet, int x, int y, ILString *str,
+							long offset, long count, int style)
+{
+	XFontStruct *fs = (XFontStruct *)fontSet;
+	ILChar *buffer = ILStringToBuffer(str) + offset;
+	XRectangle overall_ink_return;
+	XRectangle overall_logical_return;
+	int line1, line2;
+	ILChar ch;
+	long ct;
+	int posn;
+	char latinBuffer[XSHARP_BUFSIZ];
+
+	/* Draw the string using the core API */
+	XSetFont(dpy, gc, fs->fid);
+	ct = count;
+	while(ct > 0)
+	{
+		/* Convert a run of text into Latin1 */
+		posn = 0;
+		while(ct > 0 && posn < XSHARP_BUFSIZ)
+		{
+			ch = *buffer++;
+			--ct;
+			if(ch < 0x0100)
+			{
+				latinBuffer[posn++] = (char)ch;
+			}
+			else
+			{
+				latinBuffer[posn++] = (char)'?';
+			}
+		}
+
+		/* Draw the text run */
+		XDrawString(dpy, drawable, gc, x, y, latinBuffer, posn);
+
+		/* Advance past the text run so that we can draw the next one */
+		if(ct > 0)
+		{
+			x += XTextWidth(fs, latinBuffer, posn);
+		}
+	}
+
+	/* Calculate the positions of the underline and strike-out */
+	if((style & FontStyle_Underline) != 0)
+	{
+		line1 = y + 1;
+	}
+	else
+	{
+		line1 = y;
+	}
+	if((style & FontStyle_StrikeOut) != 0)
+	{
+		line2 = y - (fs->ascent / 2);
+	}
+	else
+	{
+		line2 = y;
+	}
+
+	/* Draw the underline and strike-out */
+	if(line1 != y || line2 != y)
+	{
+		XSharpTextExtentsStruct(dpy, fontSet, str, offset, count,
+				 				&overall_ink_return, &overall_logical_return);
+		XSetLineAttributes(dpy, gc, 1, LineSolid, CapNotLast, JoinMiter);
+		if(line1 != y)
+		{
+			XDrawLine(dpy, drawable, gc, x, line1,
+					  x + overall_logical_return.width, line1);
+		}
+		if(line2 != y)
+		{
+			XDrawLine(dpy, drawable, gc, x, line2,
+					  x + overall_logical_return.width, line2);
+		}
+	}
+}
+
+/*
+ * Draw a string using an Xft font.
+ */
+void XSharpDrawStringXft(Display *dpy, Drawable drawable, GC gc,
+					     void *fontSet, int x, int y,
+					     const char *str, int style, Region clipRegion,
+					     unsigned long colorValue)
+{
 #ifdef USE_XFT_EXTENSION
+
+	XRectangle overall_ink_return;
+	XRectangle overall_logical_return;
 	XftDraw *draw;
 	XftColor color;
 	XGCValues values;
-#else
-	XFontSetExtents *extents;
-#endif
 	int line1, line2;
 
-#ifdef USE_XFT_EXTENSION
-
 	/* TODO: 16-bit fonts */
 
 	/* Set up the Xft color value to draw with */
@@ -327,18 +506,10 @@ void XSharpDrawString(Display *dpy, Drawable drawable, GC gc,
 		XftDrawDestroy(draw);
 	}
 
-#else
-
-	/* Draw the string using the core API */
-	XmbDrawString(dpy, drawable, (XFontSet)fontSet, gc, x, y,
-				  str, strlen(str));
-
-#endif
-
 	/* Calculate the positions of the underline and strike-out */
 	if((style & FontStyle_Underline) != 0)
 	{
-		line1 = y + 2;
+		line1 = y + 1;
 	}
 	else
 	{
@@ -346,19 +517,7 @@ void XSharpDrawString(Display *dpy, Drawable drawable, GC gc,
 	}
 	if((style & FontStyle_StrikeOut) != 0)
 	{
-	#ifdef USE_XFT_EXTENSION
 		line2 = y + (((XftFont *)fontSet)->height / 2);
-	#else
-		extents = XExtentsOfFontSet(fontSet);
-		if(extents)
-		{
-			line2 = y + (extents->max_logical_extent.y / 2);
-		}
-		else
-		{
-			line2 = y;
-		}
-	#endif
 	}
 	else
 	{
@@ -368,8 +527,8 @@ void XSharpDrawString(Display *dpy, Drawable drawable, GC gc,
 	/* Draw the underline and strike-out */
 	if(line1 != y || line2 != y)
 	{
-		XSharpTextExtents(dpy, fontSet, str,
-				 	      &overall_ink_return, &overall_logical_return);
+		XSharpTextExtentsXft(dpy, fontSet, str,
+				 	         &overall_ink_return, &overall_logical_return);
 		XSetLineAttributes(dpy, gc, 1, LineSolid, CapNotLast, JoinMiter);
 		if(line1 != y)
 		{
@@ -382,14 +541,181 @@ void XSharpDrawString(Display *dpy, Drawable drawable, GC gc,
 					  x + overall_logical_return.width, line2);
 		}
 	}
+#endif /* USE_XFT_EXTENSION */
 }
 
 /*
- * Calculate the extent information for a string.
+ * Calculate the extent information for a string in a font set.
  */
-void XSharpTextExtents(Display *dpy, void *fontSet, const char *str,
-					   XRectangle *overall_ink_return,
-					   XRectangle *overall_logical_return)
+void XSharpTextExtentsSet(Display *dpy, void *fontSet, const char *str,
+					      XRectangle *overall_ink_return,
+					      XRectangle *overall_logical_return)
+{
+	XmbTextExtents((XFontSet)fontSet, str, strlen(str),
+			 	   overall_ink_return, overall_logical_return);
+}
+
+/*
+ * Some helper macros from the Xlib code.  Copyright by various authors
+ * and distributed under the MIT/X11 license.  See the Xlib source for
+ * further details (Xlibint.h and TextExt.c).
+ */
+#define CI_NONEXISTCHAR(cs) (((cs)->width == 0) && \
+			     (((cs)->rbearing|(cs)->lbearing| \
+			       (cs)->ascent|(cs)->descent) == 0))
+#define CI_GET_CHAR_INFO_1D(fs,col,def,cs) \
+{ \
+    cs = def; \
+    if (col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) { \
+	if (fs->per_char == NULL) { \
+	    cs = &fs->min_bounds; \
+	} else { \
+	    cs = &fs->per_char[(col - fs->min_char_or_byte2)]; \
+	    if (CI_NONEXISTCHAR(cs)) cs = def; \
+	} \
+    } \
+}
+#define CI_GET_DEFAULT_INFO_1D(fs,cs) \
+  CI_GET_CHAR_INFO_1D (fs, fs->default_char, NULL, cs)
+#define CI_GET_CHAR_INFO_2D(fs,row,col,def,cs) \
+{ \
+    cs = def; \
+    if (row >= fs->min_byte1 && row <= fs->max_byte1 && \
+	col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) { \
+	if (fs->per_char == NULL) { \
+	    cs = &fs->min_bounds; \
+	} else { \
+	    cs = &fs->per_char[((row - fs->min_byte1) * \
+			        (fs->max_char_or_byte2 - \
+				 fs->min_char_or_byte2 + 1)) + \
+			       (col - fs->min_char_or_byte2)]; \
+	    if (CI_NONEXISTCHAR(cs)) cs = def; \
+        } \
+    } \
+}
+#define CI_GET_DEFAULT_INFO_2D(fs,cs) \
+{ \
+    unsigned int r = (fs->default_char >> 8); \
+    unsigned int c = (fs->default_char & 0xff); \
+    CI_GET_CHAR_INFO_2D (fs, r, c, NULL, cs); \
+}
+#define CI_GET_ROWZERO_CHAR_INFO_2D(fs,col,def,cs) \
+{ \
+    cs = def; \
+    if (fs->min_byte1 == 0 && \
+	col >= fs->min_char_or_byte2 && col <= fs->max_char_or_byte2) { \
+	if (fs->per_char == NULL) { \
+	    cs = &fs->min_bounds; \
+	} else { \
+	    cs = &fs->per_char[(col - fs->min_char_or_byte2)]; \
+	    if (CI_NONEXISTCHAR(cs)) cs = def; \
+	} \
+    } \
+}
+
+/*
+ * Calculate the extent information for a string in a font struct.
+ *
+ * This is based on the code for the "XTextExtents" function in Xlib.
+ * We expand it inline so that we can convert from Unicode to Latin1 on
+ * the fly, rather than having to convert into a separate buffer first.
+ */
+void XSharpTextExtentsStruct(Display *dpy, void *fontSet, ILString *str,
+					         long offset, long count,
+							 XRectangle *overall_ink_return,
+					         XRectangle *overall_logical_return)
+{
+	ILChar *buffer = ILStringToBuffer(str) + offset;
+	XFontStruct *fs = (XFontStruct *)fontSet;
+    int isSingleRow = (fs->max_byte1 == 0);
+	int first = 1;
+	XCharStruct *defaultSize;
+	XCharStruct *charSize;
+	XCharStruct overall;
+	unsigned ch;
+	int temp;
+
+	/* Get the metrics for the default character */
+    if(isSingleRow)
+	{
+		CI_GET_DEFAULT_INFO_1D(fs, defaultSize);
+    }
+	else
+	{
+		CI_GET_DEFAULT_INFO_2D(fs, defaultSize);
+    }
+
+	/* Iterate over the characters and measure them */
+	overall.width = 0;
+	overall.ascent = 0;
+	overall.descent = 0;
+	overall.lbearing = 0;
+	overall.rbearing = 0;
+	while(count-- > 0)
+	{
+		ch = (unsigned)(*buffer++);
+		if(ch >= 0x0100)
+		{
+			/* Convert non-Latin1 characters into '?' */
+			ch = '?';
+		}
+		if(isSingleRow)
+		{
+	    	CI_GET_CHAR_INFO_1D(fs, ch, defaultSize, charSize);
+		}
+		else
+		{
+	    	CI_GET_ROWZERO_CHAR_INFO_2D(fs, ch, defaultSize, charSize);
+		}
+		if(charSize)
+		{
+			if(first)
+			{
+				overall = *charSize;
+				first = 0;
+			}
+			else
+			{
+				if(charSize->ascent > overall.ascent)
+				{
+					overall.ascent = charSize->ascent;
+				}
+				if(charSize->descent > overall.descent)
+				{
+					overall.descent = charSize->descent;
+				}
+				temp = overall.width + charSize->lbearing;
+				if(temp < overall.lbearing)
+				{
+					overall.lbearing = temp;
+				}
+				temp = overall.width + charSize->rbearing;
+				if(temp > overall.rbearing)
+				{
+					overall.rbearing = temp;
+				}
+				overall.width += charSize->width;
+		    }
+		}
+    }
+
+	/* Convert the raw extent information into the desired return values */
+	overall_ink_return->x = overall.lbearing;
+	overall_ink_return->y = -(overall.ascent);
+	overall_ink_return->width = overall.rbearing - overall.lbearing;
+	overall_ink_return->height = overall.ascent + overall.descent;
+	overall_logical_return->x = 0;
+	overall_logical_return->y = -(fs->ascent);
+	overall_logical_return->width = overall.width;
+	overall_logical_return->height = fs->ascent + fs->descent;
+}
+
+/*
+ * Calculate the extent information for a string in an Xft font.
+ */
+void XSharpTextExtentsXft(Display *dpy, void *fontSet, const char *str,
+					      XRectangle *overall_ink_return,
+					      XRectangle *overall_logical_return)
 {
 #ifdef USE_XFT_EXTENSION
 
@@ -408,18 +734,51 @@ void XSharpTextExtents(Display *dpy, void *fontSet, const char *str,
 	overall_logical_return->width = extents.x + extents.xOff;
 	overall_logical_return->height = extents.y + extents.yOff;
 
-#else
-	XmbTextExtents((XFontSet)fontSet, str, strlen(str),
-			 	   overall_ink_return, overall_logical_return);
 #endif
 }
 
 /*
- * Calculate the extent information for a font.
+ * Calculate the extent information for a font set.
  */
-void XSharpFontExtents(void *fontSet,
-					   XRectangle *max_ink_return,
-					   XRectangle *max_logical_return)
+void XSharpFontExtentsSet(void *fontSet,
+					      XRectangle *max_ink_return,
+					      XRectangle *max_logical_return)
+{
+	XFontSetExtents *extents;
+	extents = XExtentsOfFontSet((XFontSet)fontSet);
+	if(extents)
+	{
+		*max_ink_return = extents->max_ink_extent;
+		*max_logical_return = extents->max_logical_extent;
+	}
+}
+
+/*
+ * Calculate the extent information for a font struct.
+ */
+void XSharpFontExtentsStruct(void *fontSet,
+					         XRectangle *max_ink_return,
+					         XRectangle *max_logical_return)
+{
+	XFontStruct *fs = (XFontStruct *)fontSet;
+	max_ink_return->x = fs->min_bounds.lbearing;
+	max_ink_return->y = -(fs->max_bounds.ascent);
+	max_ink_return->width =
+		fs->max_bounds.rbearing - fs->min_bounds.lbearing;
+	max_ink_return->height =
+		fs->max_bounds.ascent + fs->max_bounds.descent;
+	max_logical_return->x = 0;
+	max_logical_return->y = -(fs->ascent);
+	max_logical_return->width = fs->max_bounds.width;
+	max_logical_return->height = fs->ascent + fs->descent;
+}
+
+/*
+ * Calculate the extent information for an Xft font.
+ */
+void XSharpFontExtentsXft(void *fontSet,
+					      XRectangle *max_ink_return,
+					      XRectangle *max_logical_return)
 {
 #ifdef USE_XFT_EXTENSION
 
@@ -431,16 +790,6 @@ void XSharpFontExtents(void *fontSet,
 								 ((XftFont *)fontSet)->descent;
 	*max_ink_return = *max_logical_return;
 
-#else
-
-	XFontSetExtents *extents;
-	extents = XExtentsOfFontSet((XFontSet)fontSet);
-	if(extents)
-	{
-		*max_ink_return = extents->max_ink_extent;
-		*max_logical_return = extents->max_logical_extent;
-	}
-
 #endif
 }
 
@@ -715,6 +1064,60 @@ static void Write_Direct(Display *dpy, Colormap colormap, XImage *image,
 	}
 }
 
+/*
+ * Write RGB data directly to an XImage, and perform an endian byteswap.
+ */
+static void Write_DirectSwap(Display *dpy, Colormap colormap, XImage *image,
+						     int line, unsigned long *input, int num)
+{
+	int column;
+	if(sizeof(unsigned int) == 4)
+	{
+		unsigned int *output = (unsigned int *)
+			(image->data + line * image->bytes_per_line);
+		for(column = 0; column < num; ++column)
+		{
+			unsigned int value = (unsigned int)(*input++);
+			value = ((value & 0x00FF0000) >> 8) |
+					((value & 0x0000FF00) << 8) |
+					((value & 0x000000FF) << 24);
+			*output++ = value;
+		}
+	}
+	else
+	{
+		unsigned long *output = (unsigned long *)
+			(image->data + line * image->bytes_per_line);
+		for(column = 0; column < num; ++column)
+		{
+			unsigned int value = (unsigned int)(*input++);
+			value = ((value & 0x00FF0000) >> 8) |
+					((value & 0x0000FF00) << 8) |
+					((value & 0x000000FF) << 24);
+			*output++ = (unsigned long)value;
+		}
+	}
+}
+
+/*
+ * Write RGB data to a 16 bits per pixel XImage structure.
+ */
+static void Write_16bpp(Display *dpy, Colormap colormap, XImage *image,
+					    int line, unsigned long *input, int num)
+{
+	int column;
+	unsigned short *output = (unsigned short *)
+		(image->data + line * image->bytes_per_line);
+	for(column = 0; column < num; ++column)
+	{
+		unsigned int value = (unsigned int)(*input++);
+		value = ((value & 0x00F80000) >> 8) |
+				((value & 0x0000FC00) >> 5) |
+				((value & 0x000000F8) >> 3);
+		*output++ = (unsigned short)value;
+	}
+}
+
 /*
  * Default writing routine, doing per-pixel color allocations.
  */
@@ -745,12 +1148,54 @@ typedef void (*WriteFunc)(Display *dpy, Colormap colormap, XImage *image,
 						  int len, unsigned long *input, int num);
 static WriteFunc GetWriteFunc(XImage *image)
 {
+	/* Determine if the client machine is little-endian or big-endian */
+	union
+	{
+		short volatile value;
+		unsigned char volatile buf[2];
+	} un;
+	int littleEndian;
+	un.value = 0x0102;
+	littleEndian = (un.buf[0] == 0x02);
+
+	/* Check for well-known special cases */
 	if(image->depth == 24 && image->red_mask == 0x00FF0000 &&
 	   image->green_mask == 0x0000FF00 && image->blue_mask == 0x000000FF &&
 	   image->byte_order == LSBFirst && image->bitmap_bit_order == LSBFirst)
 	{
-		return Write_Direct;
+		if(littleEndian)
+		{
+			return Write_Direct;
+		}
+		else
+		{
+			return Write_DirectSwap;
+		}
 	}
+	if(image->depth == 24 && image->red_mask == 0x00FF0000 &&
+	   image->green_mask == 0x0000FF00 && image->blue_mask == 0x000000FF &&
+	   image->byte_order == MSBFirst && image->bitmap_bit_order == MSBFirst)
+	{
+		if(littleEndian)
+		{
+			return Write_DirectSwap;
+		}
+		else
+		{
+			return Write_Direct;
+		}
+	}
+	if(image->depth == 16 && image->red_mask == 0x0000F800 &&
+	   image->green_mask == 0x000007E0 && image->blue_mask == 0x0000001F &&
+	   image->byte_order == LSBFirst && image->bitmap_bit_order == LSBFirst)
+	{
+		if(littleEndian)
+		{
+			return Write_16bpp;
+		}
+	}
+
+	/* Fall back to the generic writer, using XPutPixel */
 	return Write_Default;
 }
 
@@ -933,7 +1378,7 @@ XImage *XSharpCreateImageFromDIB(Screen *screen, int width, int height,
 			if(pixelFormat == PF_Format1bppIndexed)
 			{
 				/* Source image has one bit per pixel */
-				for(column = 0; column < width; column += 2)
+				for(column = 0; column < width; ++column)
 				{
 					if((temp1[column / 8] & (0x80 >> (column % 8))) != 0)
 					{
@@ -1094,43 +1539,185 @@ void XSharpGetRegionRect(Region region, int index, XRectangle *rect)
 	rect->height = box->y2 - box->y1;
 }
 
+/*
+ * Send a close request to a window.  Tries to send "WM_DELETE_WINDOW",
+ * and if that doesn't work, sends "XKillClent" instead.
+ */
+void XSharpSendClose(Display *dpy, Window window)
+{
+	Atom wmDeleteWindow = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
+	Atom *protocols;
+	int num_protocols, posn;
+	int use_kill = 1;
+
+	/* Get the protocols that are supported by the client window */
+	protocols = 0;
+	num_protocols = 0;
+	if(XGetWMProtocols(dpy, window, &protocols, &num_protocols))
+	{
+		for(posn = 0; posn < num_protocols; ++posn)
+		{
+			if(protocols[posn] == wmDeleteWindow)
+			{
+				use_kill = 0;
+				break;
+			}
+		}
+		if(protocols)
+		{
+			XFree(protocols);
+		}
+	}
+
+	/* Send the close request to the client application */
+	if(use_kill)
+	{
+		XKillClient(dpy, (XID)window);
+	}
+	else
+	{
+		XEvent msg;
+		memset(&msg, 0, sizeof(msg));
+		msg.xclient.type = ClientMessage;
+		msg.xclient.window = window;
+		msg.xclient.message_type = XInternAtom(dpy, "WM_PROTOCOLS", False);
+		msg.xclient.format = 32;
+		msg.xclient.data.l[0] = wmDeleteWindow;
+		XSendEvent(dpy, window, False, NoEventMask, &msg);
+	}
+}
+
+/*
+ * Send a wakeup message to a window.
+ */
+void XSharpSendWakeup(Display *dpy, Window window)
+{
+	Atom wakeupAtom = XInternAtom(dpy, "_XSHARP_WAKEUP", False);
+	XEvent msg;
+	memset(&msg, 0, sizeof(msg));
+	msg.xclient.type = ClientMessage;
+	msg.xclient.window = window;
+	msg.xclient.message_type = wakeupAtom;
+	msg.xclient.format = 32;
+	XSendEvent(dpy, window, False, NoEventMask, &msg);
+}
+
 #else /* X_DISPLAY_MISSING || !HAVE_SELECT */
 
+int XSharpSupportPresent(void)
+{
+	return 0;
+}
+
 int XNextEventWithTimeout(void *dpy, void *event, int timeout)
 {
 	/* Nothing to do here */
 	return -1;
 }
 
-void *XSharpCreateFont(void *dpy, const char *family, int pointSize, int style)
+int XSharpUseXft()
 {
 	/* Nothing to do here */
 	return 0;
 }
 
-void XSharpFreeFont(void *dpy, void *fontSet)
+void *XSharpCreateFontSet(void *dpy, const char *family,
+						  int pointSize, int style)
+{
+	/* Nothing to do here */
+	return 0;
+}
+
+void XSharpFreeFontSet(void *dpy, void *fontSet)
 {
 	/* Nothing to do here */
 }
 
-void XSharpDrawString(void *dpy, unsigned long drawable, void *gc,
-					  void *fontSet, int x, int y,
-					  const char *str, int style, void *clipRegion,
-					  unsigned long colorValue)
+void XSharpDrawStringSet(void *dpy, unsigned long drawable, void *gc,
+					     void *fontSet, int x, int y,
+					     const char *str, int style)
 {
 	/* Nothing to do here */
 }
 
-void XSharpTextExtents(void *dpy, void *fontSet, const char *str,
-					   void *overall_ink_return,
-					   void *overall_logical_return)
+void XSharpTextExtentsSet(void *dpy, void *fontSet, const char *str,
+					      void *overall_ink_return,
+					      void *overall_logical_return)
 {
 	/* Nothing to do here */
 }
 
-void XSharpFontExtents(void *fontSet,
-					   void *max_ink_return,
-					   void *max_logical_return)
+void XSharpFontExtentsSet(void *fontSet,
+					      void *max_ink_return,
+					      void *max_logical_return)
+{
+	/* Nothing to do here */
+}
+
+void *XSharpCreateFontStruct(void *dpy, const char *family,
+							 int pointSize, int style)
+{
+	/* Nothing to do here */
+	return 0;
+}
+
+void XSharpFreeFontStruct(void *dpy, void *fontSet)
+{
+	/* Nothing to do here */
+}
+
+void XSharpDrawStringStruct(void *dpy, unsigned long drawable, void *gc,
+					        void *fontSet, int x, int y,
+					        void *str, long offset, long count, int style)
+{
+	/* Nothing to do here */
+}
+
+void XSharpTextExtentsStruct(void *dpy, void *fontSet,
+							 void *str, long offset, long count,
+					         void *overall_ink_return,
+					         void *overall_logical_return)
+{
+	/* Nothing to do here */
+}
+
+void XSharpFontExtentsStruct(void *fontSet,
+					         void *max_ink_return,
+					         void *max_logical_return)
+{
+	/* Nothing to do here */
+}
+
+void *XSharpCreateFontXft(void *dpy, const char *family,
+						  int pointSize, int style)
+{
+	/* Nothing to do here */
+	return 0;
+}
+
+void XSharpFreeFontXft(void *dpy, void *fontSet)
+{
+	/* Nothing to do here */
+}
+
+void XSharpDrawStringXft(void *dpy, unsigned long drawable, void *gc,
+					     void *fontSet, int x, int y,
+					     const char *str, int style, void *clipRegion,
+					     unsigned long colorValue)
+{
+	/* Nothing to do here */
+}
+
+void XSharpTextExtentsXft(void *dpy, void *fontSet, const char *str,
+					      void *overall_ink_return,
+					      void *overall_logical_return)
+{
+	/* Nothing to do here */
+}
+
+void XSharpFontExtentsXft(void *fontSet,
+					      void *max_ink_return,
+					      void *max_logical_return)
 {
 	/* Nothing to do here */
 }
@@ -1177,4 +1764,14 @@ void XSharpGetRegionRect(void *region, int index, void *rect)
 	/* Nothing to do here */
 }
 
+void XSharpSendClose(void *dpy, unsigned long window)
+{
+	/* Nothing to do here */
+}
+
+void XSharpSendWakeup(void *dpy, unsigned long window)
+{
+	/* Nothing to do here */
+}
+
 #endif /* X_DISPLAY_MISSING || !HAVE_SELECT */
