@@ -24,21 +24,26 @@ namespace System
 [Flags]
 public enum AttributeTargets
 {
-	Assembly		= 0x0001,
-	Module			= 0x0002,
-	Class			= 0x0004,
-	Struct			= 0x0008,
-	Enum			= 0x0010,
-	Constructor		= 0x0020,
-	Method			= 0x0040,
-	Property		= 0x0080,
-	Field			= 0x0100,
-	Event			= 0x0200,
-	Interface		= 0x0400,
-	Parameter		= 0x0800,
-	Delegate		= 0x1000,
-	ReturnValue		= 0x2000,
-	All				= 0x3FFF
+	Assembly		 = 0x0001,
+	Module			 = 0x0002,
+	Class			 = 0x0004,
+	Struct			 = 0x0008,
+	Enum			 = 0x0010,
+	Constructor		 = 0x0020,
+	Method			 = 0x0040,
+	Property		 = 0x0080,
+	Field			 = 0x0100,
+	Event			 = 0x0200,
+	Interface		 = 0x0400,
+	Parameter		 = 0x0800,
+	Delegate		 = 0x1000,
+	ReturnValue		 = 0x2000,
+#if CONFIG_GENERICS
+	GenericParameter = 0x4000,
+	All				 = 0x7FFF
+#else
+	All				 = 0x3FFF
+#endif
 
 }; // enum AttributeTargets
 
